"""
Author: Wenyu Ouyang
Date: 2021-12-05 11:21:58
LastEditTime: 2024-05-23 15:21:17
LastEditors: Wenyu Ouyang
Description: Main function for training and testing
FilePath: \torchhydro\torchhydro\trainers\trainer.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import copy
from datetime import datetime
import os
from pathlib import Path
import random

import numpy as np
from typing import Dict
import pandas as pd
from sklearn.model_selection import KFold, TimeSeriesSplit
import torch
from torchhydro.trainers.deep_hydro import model_type_dict
from torchhydro.trainers.resulter import Resulter


def set_random_seed(seed):
    """
    Set a random seed to guarantee the reproducibility

    Parameters
    ----------
    seed
        a number

    Returns
    -------
    None
    """
    # print("Random seed:", seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def train_and_evaluate(cfgs: Dict):
    """
    Function to train and test a Model

    Parameters
    ----------
    cfgs
        Dictionary containing all configs needed to run the model

    Returns
    -------
    None
    """
    random_seed = cfgs["training_cfgs"]["random_seed"]
    set_random_seed(random_seed)
    resulter = Resulter(cfgs)
    deephydro = _get_deep_hydro(cfgs)
    if cfgs["training_cfgs"]["train_mode"] and (
        (
            deephydro.weight_path is not None
            and deephydro.cfgs["model_cfgs"]["continue_train"]
        )
        or (deephydro.weight_path is None)
    ):
        deephydro.model_train()
    preds, obss = deephydro.model_evaluate()
    resulter.save_cfg(deephydro.cfgs)
    resulter.save_result(preds, obss)
    resulter.eval_result(preds, obss)


def _get_deep_hydro(cfgs):
    model_type = cfgs["model_cfgs"]["model_type"]
    return model_type_dict[model_type](cfgs)


def _update_cfg_with_1ensembleitem(cfg, key, value):
    """update a dict with key and value

    Parameters
    ----------
    my_dict : _type_
        _description_
    key : _type_
        _description_
    value : _type_
        _description_
    """
    new_cfg = copy.deepcopy(cfg)
    if key == "kfold":
        new_cfg["data_cfgs"]["t_range_train"] = value[0]
        # although cross_validation contain the name "valid",
        # it is actually the test period for our model's evaluating
        new_cfg["data_cfgs"]["t_range_valid"] = None
        new_cfg["data_cfgs"]["t_range_test"] = value[1]
    elif key == "batch_sizes":
        new_cfg["training_cfgs"]["batch_size"] = value
        new_cfg["data_cfgs"]["batch_size"] = value
    elif key == "seeds":
        new_cfg["training_cfgs"]["random_seed"] = value
    elif key == "expdir":
        project_dir = new_cfg["data_cfgs"]["test_path"]
        project_path = Path(project_dir)
        subset = project_path.parent.name
        subexp = f"{project_path.name}_{value}"
        new_cfg["data_cfgs"]["test_path"] = os.path.join(
            project_path.parent.parent, subset, subexp
        )
    else:
        raise ValueError(f"key {key} is not supported")
    return new_cfg


def _create_kfold_periods(train_period, test_period, kfold):
    """
    Create k folds from the complete time period defined by the earliest start date and latest end date
    among train, valid, and test periods, ignoring any period that is None or has NaN values.

    Parameters:
    - train_period: List with 2 elements [start_date, end_date] for training period.
    - test_period: List with 2 elements [start_date, end_date] for testing period.
    - kfold: Number of folds to split the data into.

    Returns:
    - A list of k elements, each element is a tuple containing two lists: train period and test period.
    """

    # Collect periods, ignoring None or NaN
    periods = [train_period, test_period]
    periods = [p for p in periods if p and not pd.isna(p[0]) and not pd.isna(p[1])]

    # Convert string dates to datetime objects and find the earliest start and latest end dates
    dates = [
        datetime.strptime(date, "%Y-%m-%d") for period in periods for date in period
    ]
    start_date = min(dates)
    end_date = max(dates)

    # Create a continuous period
    full_period = pd.date_range(start=start_date, end=end_date)
    periods = np.array(range(len(full_period)))

    # Apply KFold
    # kf = KFold(n_splits=kfold)
    kf = TimeSeriesSplit(n_splits=kfold)
    folds = []
    for train_index, test_index in kf.split(periods):
        train_start = full_period[train_index[0]].strftime("%Y-%m-%d")
        train_end = full_period[train_index[-1]].strftime("%Y-%m-%d")
        test_start = full_period[test_index[0]].strftime("%Y-%m-%d")
        test_end = full_period[test_index[-1]].strftime("%Y-%m-%d")
        folds.append(([train_start, train_end], [test_start, test_end]))

    return folds


def _create_kfold_discontinuous_periods(train_period, test_period, kfold):
    periods = train_period + test_period
    periods = sorted(periods, key=lambda x: x[0])
    cross_validation_sets = []

    for i in range(kfold):
        valid = [periods[i]]
        train = [p for j, p in enumerate(periods) if j != i]
        cross_validation_sets.append((train, valid))
    return cross_validation_sets


def _nested_loop_train_and_evaluate(keys, index, my_dict, update_dict, perform_count=0):
    """a recursive function to update the update_dict and perform train_and_evaluate

    Parameters
    ----------
    keys : list
        a list of keys
    index : int
        the loop index
    my_dict : dict
        the dict we want to loop
    update_dict : dict
        the dict we want to update
    perform_count : int, optional
        a counter for naming different experiments, by default 0
    """
    if index == len(keys):
        return perform_count
    current_key = keys[index]
    for value in my_dict[current_key]:
        # update the update_dict
        cfg = _update_cfg_with_1ensembleitem(update_dict, current_key, value)
        # for final key, perform train and evaluate
        if index == len(keys) - 1:
            cfg = _update_cfg_with_1ensembleitem(cfg, "expdir", perform_count)
            train_and_evaluate(cfg)
            # print(perform_count)
            perform_count += 1
        # recursive
        perform_count = _nested_loop_train_and_evaluate(
            keys, index + 1, my_dict, cfg, perform_count
        )
    return perform_count


def _trans_kfold_to_periods(update_dict, ensemble_items, current_key="kfold"):
    # set train and test period
    train_period = update_dict["data_cfgs"]["t_range_train"]
    test_period = update_dict["data_cfgs"]["t_range_test"]
    kfold = ensemble_items[current_key]
    if ensemble_items["kfold_continuous"]:
        kfold_periods = _create_kfold_periods(train_period, test_period, kfold)
    else:
        kfold_periods = _create_kfold_discontinuous_periods(
            train_period, test_period, kfold
        )
    ensemble_items[current_key] = kfold_periods


def ensemble_train_and_evaluate(cfgs: Dict):
    """
    Function to train and test for ensemble models

    Parameters
    ----------
    cfgs
        Dictionary containing all configs needed to run the model

    Returns
    -------
    None
    """
    # for basins and models
    ensemble = cfgs["training_cfgs"]["ensemble"]
    if not ensemble:
        raise ValueError(
            "ensemble should be True, otherwise should use train_and_evaluate rather than ensemble_train_and_evaluate"
        )
    ensemble_items = cfgs["training_cfgs"]["ensemble_items"]
    number_of_items = len(ensemble_items)
    if number_of_items == 0:
        raise ValueError("ensemble_items should not be empty")
    keys_list = list(ensemble_items.keys())
    if "kfold" in keys_list:
        _trans_kfold_to_periods(cfgs, ensemble_items, "kfold")
    _nested_loop_train_and_evaluate(keys_list, 0, ensemble_items, cfgs)



"""
Author: Xinzhuo Wu
Date: 2023-09-30 1:20:18
LastEditTime: 2024-02-13 13:18:21
LastEditors: Wenyu Ouyang
Description: Test a full training and evaluating process with Spp_Lstm
FilePath: \torchhydro\tests\test_spp_lstm.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import pytest
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate, ensemble_train_and_evaluate
import warnings

warnings.filterwarnings("ignore")


@pytest.fixture()
def config():
    project_name = "test_spp_lstm/ex1"
    config_data = default_config_file()
    args = cmd(
        sub=project_name,
        source="HydroGrid",
        source_path=[
            {
                "gpm": "basins-origin/hour_data/1h/grid_data/grid_gpm_data",
                "gfs": "basins-origin/hour_data/1h/grid_data/grid_gfs_data",
                "smap": "basins-origin/hour_data/1h/grid_data/grid_smap_data",
                "target": "basins-origin/hour_data/1h/mean_data/mean_data_target",
                "attributes": "basins-origin/attributes.nc",
            }
        ],
        ctx=[2],
        model_name="SPPLSTM2",
        model_hyperparam={  # p代表降水+GFS气象网络分支，S代表土壤属性网络分支
            "forecast_history": 168,
            "forecast_length": 24,
            "p_n_output": 1,
            "p_n_hidden_states": 60,
            "p_dropout": 0.25,
            "p_in_channels": 1,  # 1 or 9 卷积层输入通道数，等于len(var_t[0]+len(var_t[1]))，即降水与GFS气象属性的数量
            "p_out_channels": 8,  # 输出通道数，值越大，模型越复杂，建议比in_channels大
            "len_c": 15,  # 需要与len(var_c)相等,即如果不用流域属性，则设置为0
            "s_forecast_history": None,  # 土壤属性所用历史时段，比如只能获取两天半前的数据，则为108,目前只支持108
            "s_n_output": 1,
            "s_n_hidden_states": 60,
            "s_dropout": 0.25,
            "s_in_channels": 8,  # 卷积层输入通道数，等于len(var_t[2])，即土壤属性的数量
            "s_out_channels": 8,
        },
        gage_id=["02051500", "21401550"],
        batch_size=256,
        var_t=[  # var_t[0]为["gpm_tp"], var_t[1]为GFS, var_t[2]为土壤含水量，无则["None"]
            ["gpm_tp"],
            [
                "None",
                # "gfs_tp",
                # "dswrf",
                # "pwat",
                # "2r",
                # "2sh",
                # "2t",
                # "tcc",
                # "10u",
                # "10v",
            ],
            [
                "None",
            ],
        ],
        var_c=[
            "area",  # 面积
            "ele_mt_smn",  # 海拔(空间平均)
            "slp_dg_sav",  # 地形坡度 (空间平均)
            "sgr_dk_sav",  # 河流坡度 (平均)
            "for_pc_sse",  # 森林覆盖率
            "glc_cl_smj",  # 土地覆盖类型
            "run_mm_syr",  # 陆面径流 (流域径流的空间平均值)
            "inu_pc_slt",  # 淹没范围 (长期最大)
            "cmi_ix_syr",  # 气候湿度指数
            "aet_mm_syr",  # 实际蒸散发 (年平均)
            "snw_pc_syr",  # 雪盖范围 (年平均)
            "swc_pc_syr",  # 土壤水含量
            "gwt_cm_sav",  # 地下水位深度
            "cly_pc_sav",  # 土壤中的黏土、粉砂、砂粒含量
            "dor_pc_pva",  # 调节程度
        ],
        var_out=["streamflow"],
        dataset="GridDataset",
        sampler="HydroSampler",
        scaler="MutiBasinScaler",
        train_epoch=1,
        save_epoch=1,
        te=1,
        train_period=[
            ("2017-07-08", "2017-09-29"),
            ("2018-07-08", "2018-09-29"),
            ("2019-07-08", "2019-09-29"),
            ("2020-07-08", "2020-09-29"),
        ],
        test_period=[
            ("2021-07-08", "2021-09-29"),
        ],
        valid_period=[
            ("2021-07-08", "2021-09-29"),
        ],
        loss_func="RMSESum",  # NSELoss、RMSESum、MAPELoss、MASELoss、MAELoss 可选
        opt="Adam",
        lr_scheduler={
            "lr": 0.001,
            "lr_factor": 0.96,
        },  # 初始化学习率(第1轮开始即使用1e-3)，可强制指定某一轮使用的学习率； 学习率更新权重
        which_first_tensor="sequence",
        early_stopping=True,
        patience=4,  # 连续n次valid loss不下降，则停止训练，与early_stopping配合使用
        rolling=False,  # evaluate 不采用滚动预测
        ensemble=True,  # 交叉验证
        ensemble_items={
            "kfold": 5,  # exi_0即17年验证,...exi_4即21年验证
            "batch_sizes": [256],
        },
    )
    update_cfg(config_data, args)
    return config_data


def test_spp_lstm(config):
    train_and_evaluate(config)
    # ensemble_train_and_evaluate(config)



from torchhydro.configs.config import default_config_file, cmd


def test_default_config_file():
    config = default_config_file()
    assert isinstance(config, dict)
    assert "model_cfgs" in config
    assert "data_cfgs" in config
    assert "training_cfgs" in config
    assert "evaluation_cfgs" in config


def test_cmd_default_args():
    # Call cmd() directly, which will return the parsed arguments
    args = cmd()

    # Assert default values
    assert args.sub is None
    assert args.source_cfgs is None
    assert args.scaler is None
    assert args.scaler_params is None
    assert args.dataset is None
    assert args.sampler is None
    assert args.fl_sample is None
    assert args.fl_num_users is None
    assert args.fl_local_ep is None
    assert args.fl_local_bs is None
    assert args.fl_frac is None
    assert args.master_addr is None
    assert args.port is None
    assert args.ctx is None
    assert args.rs is None
    assert args.train_mode is None
    assert args.train_epoch is None
    assert args.save_epoch is None
    assert args.save_iter is None
    assert args.loss_func is None
    assert args.loss_param is None
    assert args.train_period is None
    assert args.valid_period is None
    assert args.test_period is None
    assert args.batch_size is None
    assert args.dropout is None
    assert args.warmup_length == 0



import os
import numpy as np
import pytest

import torch

from torchhydro import SETTING
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate
from torchhydro.models.dpl4xaj import DplLstmXaj
from torchhydro.models.kernel_conv import uh_conv, uh_gamma
from torchhydro.models.dpl4xaj_nn4et import DplLstmNnModuleXaj


@pytest.fixture()
def device():
    return "cuda:0" if torch.cuda.is_available() else "cpu"


@pytest.fixture()
def dpl(device):
    dpl_ = DplLstmXaj(5, 15, 64, kernel_size=15, warmup_length=5)
    return dpl_.to(device)


@pytest.fixture()
def dpl_eh(device):
    dpl_ = DplLstmXaj(
        5,
        15,
        64,
        kernel_size=15,
        warmup_length=5,
        source_book="EH",
        source_type="sources",
    )
    return dpl_.to(device)


@pytest.fixture()
def dpl_eh5mm(device):
    dpl_ = DplLstmXaj(
        5,
        15,
        64,
        kernel_size=15,
        warmup_length=5,
        source_book="EH",
        source_type="sources5mm",
    )
    return dpl_.to(device)


@pytest.fixture()
def dpl_nnmodule(device):
    dpl_ = DplLstmNnModuleXaj(
        5, 15, 64, kernel_size=15, warmup_length=5, et_output=1, param_var_index=[]
    )
    return dpl_.to(device)


def test_dpl_lstm_xaj(device, dpl):
    # sequence-first tensor: time_sequence, batch, feature_size (assume that they are p, pet, srad, tmax, tmin)
    x = torch.rand(20, 10, 5).to(device)
    z = torch.rand(20, 10, 5).to(device)
    qe = dpl(x, z)
    assert len(qe.shape) == 3
    assert qe.shape == (15, 10, 2)
    qe.backward(torch.ones_like(qe))
    assert isinstance(qe, torch.Tensor)


def test_dpl_lstm_xaj_eh(device, dpl_eh):
    # sequence-first tensor: time_sequence, batch, feature_size (assume that they are p, pet, srad, tmax, tmin)
    x = torch.rand(20, 10, 5).to(device)
    z = torch.rand(20, 10, 5).to(device)
    qe = dpl_eh(x, z)
    assert len(qe.shape) == 3
    assert qe.shape == (15, 10, 2)
    qe.backward(torch.ones_like(qe))
    assert isinstance(qe, torch.Tensor)


def test_dpl_lstm_xaj_eh5mm(device, dpl_eh5mm):
    # sequence-first tensor: time_sequence, batch, feature_size (assume that they are p, pet, srad, tmax, tmin)
    x = torch.rand(20, 10, 5).to(device)
    z = torch.rand(20, 10, 5).to(device)
    qe = dpl_eh5mm(x, z)
    assert len(qe.shape) == 3
    assert qe.shape == (15, 10, 2)
    qe.backward(torch.ones_like(qe))
    assert isinstance(qe, torch.Tensor)


def test_uh_gamma():
    # batch = 10
    tempa = torch.Tensor(np.full(10, [2.5]))
    tempb = torch.Tensor(np.full(10, [3.5]))
    # repeat for 20 periods and add one dim as feature: time_seq, batch, feature
    routa = tempa.repeat(20, 1).unsqueeze(-1)
    routb = tempb.repeat(20, 1).unsqueeze(-1)
    uh = uh_gamma(routa, routb, len_uh=15)
    np.testing.assert_almost_equal(
        uh.numpy()[:, 0, :],
        np.array(
            [
                [0.0069],
                [0.0314],
                [0.0553],
                [0.0738],
                [0.0860],
                [0.0923],
                [0.0939],
                [0.0919],
                [0.0875],
                [0.0814],
                [0.0744],
                [0.0670],
                [0.0597],
                [0.0525],
                [0.0459],
            ]
        ),
        decimal=3,
    )


def test_uh():
    uh_from_gamma = torch.full((5, 3, 1), 1.0)
    # uh_from_gamma = torch.Tensor(np.arange(15).reshape(5, 3, 1))
    rf = torch.Tensor(np.arange(30).reshape(10, 3, 1) / 100)
    qs = uh_conv(rf, uh_from_gamma)
    np.testing.assert_almost_equal(
        np.array(
            [
                [0.0000, 0.0100, 0.0200],
                [0.0300, 0.0500, 0.0700],
                [0.0900, 0.1200, 0.1500],
                [0.1800, 0.2200, 0.2600],
                [0.3000, 0.3500, 0.4000],
                [0.4500, 0.5000, 0.5500],
                [0.6000, 0.6500, 0.7000],
                [0.7500, 0.8000, 0.8500],
                [0.9000, 0.9500, 1.0000],
                [1.0500, 1.1000, 1.1500],
            ]
        ),
        qs.numpy()[:, :, 0],
        decimal=3,
    )


def test_dpl_lstm_xaj_nnmodule(device, dpl_nnmodule):
    # sequence-first tensor: time_sequence, batch, feature_size (assume that they are p, pet, srad, tmax, tmin)
    x = torch.rand(20, 10, 5).to(device)
    z = torch.rand(20, 10, 5).to(device)
    qe = dpl_nnmodule(x, z)
    assert len(qe.shape) == 3
    assert qe.shape == (15, 10, 2)
    assert isinstance(qe, torch.Tensor)


def test_train_evaluate_dpl(dpl_args, config_data):
    update_cfg(config_data, dpl_args)
    train_and_evaluate(config_data)


@pytest.fixture()
def dpl_selfmadehydrodataset_args():
    project_name = os.path.join("test_camels", "expdpl61561201")
    train_period = ["2014-10-01", "2018-10-01"]
    valid_period = ["2017-10-01", "2021-10-01"]
    # valid_period = None
    test_period = ["2017-10-01", "2021-10-01"]
    return cmd(
        sub=project_name,
        source_cfgs={
            "source_name": "selfmadehydrodataset",
            "source_path": SETTING["local_data_path"]["datasets-interim"],
            "other_settings": {"time_unit": ["1D"]},
        },
        model_type="MTL",
        ctx=[0],
        # model_name="DplLstmXaj",
        # model_name="DplAttrXaj",
        model_name="DplNnModuleXaj",
        model_hyperparam={
            "n_input_features": 6,
            # "n_input_features": 19,
            "n_output_features": 15,
            "n_hidden_states": 64,
            "kernel_size": 15,
            "warmup_length": 365,
            "param_limit_func": "clamp",
            "param_test_way": "final",
            "source_book": "HF",
            "source_type": "sources",
            "et_output": 1,
            "param_var_index": [],
        },
        # loss_func="RMSESum",
        loss_func="MultiOutLoss",
        loss_param={
            "loss_funcs": "RMSESum",
            "data_gap": [0, 0],
            "device": [0],
            "item_weight": [1, 0],
            "limit_part": [1],
        },
        dataset="DplDataset",
        scaler="DapengScaler",
        scaler_params={
            "prcp_norm_cols": [
                "streamflow",
            ],
            "gamma_norm_cols": [
                "total_precipitation_hourly",
                "potential_evaporation_hourly",
            ],
            "pbm_norm": True,
        },
        gage_id=[
            # "camels_01013500",
            # "camels_01022500",
            # "camels_01030500",
            # "camels_01031500",
            # "camels_01047000",
            # "camels_01052500",
            # "camels_01054200",
            # "camels_01055000",
            # "camels_01057000",
            # "camels_01170100",
            "changdian_61561"
        ],
        train_period=train_period,
        valid_period=valid_period,
        test_period=test_period,
        batch_size=300,
        forecast_history=0,
        forecast_length=365,
        var_t=[
            # although the name is hourly, it might be daily according to your choice
            "total_precipitation_hourly",
            "potential_evaporation_hourly",
            "snow_depth_water_equivalent",
            "snowfall_hourly",
            "dewpoint_temperature_2m",
            "temperature_2m",
        ],
        var_c=[
            # "sgr_dk_sav",
            # "pet_mm_syr",
            # "slp_dg_sav",
            # "for_pc_sse",
            # "pre_mm_syr",
            # "slt_pc_sav",
            # "swc_pc_syr",
            # "soc_th_sav",
            # "cly_pc_sav",
            # "ari_ix_sav",
            # "snd_pc_sav",
            # "ele_mt_sav",
            # "area",
            # "tmp_dc_syr",
            # "crp_pc_sse",
            # "lit_cl_smj",
            # "wet_cl_smj",
            # "snw_pc_syr",
            # "glc_cl_smj",
        ],
        # NOTE: although we set total_evaporation_hourly as output, it is not used in the training process
        var_out=["streamflow", "total_evaporation_hourly"],
        n_output=2,
        # TODO: if chose "mean", metric results' format is different, this should be refactored
        fill_nan=["no", "no"],
        target_as_input=0,
        constant_only=0,
        # train_epoch=100,
        train_epoch=2,
        save_epoch=10,
        model_loader={
            "load_way": "specified",
            # "test_epoch": 100,
            "test_epoch": 2,
        },
        warmup_length=365,
        opt="Adadelta",
        which_first_tensor="sequence",
        # train_mode=0,
        # weight_path="C:\\Users\\wenyu\\code\\torchhydro\\results\\test_camels\\expdpl61561201\\10_September_202402_32PM_model.pth",
        # continue_train=0,
    )


def test_dpl_selfmadehydrodataset(dpl_selfmadehydrodataset_args):
    cfg = default_config_file()
    update_cfg(cfg, dpl_selfmadehydrodataset_args)
    train_and_evaluate(cfg)
    print("All processes are finished!")



"""
Author: Wenyu Ouyang
Date: 2023-07-31 08:40:43
LastEditTime: 2024-05-27 15:53:43
LastEditors: Wenyu Ouyang
Description: Test some functions for dataset
FilePath: \torchhydro\tests\test_sampler.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
import hydrodataset as hds
from hydrodataset.caravan import Caravan
from torch.utils.data import Dataset

from torchhydro import SETTING
from torchhydro.datasets.sampler import KuaiSampler


class SimpleDataset(Dataset):
    def __len__(self):
        return 1000

    def __getitem__(self, idx):
        return idx


@pytest.fixture
def dataset():
    return SimpleDataset()


@pytest.fixture
def sampler(dataset):
    return KuaiSampler(
        dataset,
        batch_size=10,
        warmup_length=5,
        rho_horizon=20,
        ngrid=10,
        nt=365,
    )


def test_sampler_length(sampler):
    num_samples = 810
    assert len(sampler) == num_samples, f"Expected {num_samples} but got {len(sampler)}"


def test_sampler_within_range(sampler, dataset):
    for idx in sampler:
        assert 0 <= idx < len(dataset), f"Index {idx} out of bounds"


def test_sampler_randomness(sampler):
    first_epoch_samples = list(sampler)
    second_epoch_samples = list(sampler)

    # a simple check for two random samples
    # the probability of two samples being the same is very small
    assert (
        first_epoch_samples != second_epoch_samples
    ), "Expected different samples for different epochs"


def test_cache_file():
    """
    Test whether the cache file is generated correctly
    """
    camels_dir = os.path.join(
        SETTING["local_data_path"]["datasets-origin"], "camels", "camels_us"
    )
    camels_us = hds.Camels(camels_dir)
    camels_us.cache_xrdataset()


def test_cache_caravan():
    """
    Test whether the cache file is generated correctly
    """
    caravan = Caravan(
        os.path.join(SETTING["local_data_path"]["datasets-origin"], "caravan")
    )
    caravan.cache_xrdataset()



"""
Author: Wenyu Ouyang
Date: 2024-09-15 11:23:28
LastEditTime: 2024-09-15 17:07:04
LastEditors: Wenyu Ouyang
Description: Test the Resulter class
FilePath: \torchhydro\tests\test_resulter.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
import xarray as xr
import numpy as np
from torchhydro.trainers.resulter import Resulter


@pytest.fixture
def mock_cfgs(tmpdir):
    return {
        "data_cfgs": {
            "test_path": str(tmpdir),
            "target_cols": ["streamflow", "total_evaporation_hourly"],
        },
        "evaluation_cfgs": {
            "model_loader": {"load_way": "latest"},
            "metrics": ["RMSE", "NSE"],
        },
        "training_cfgs": {"epochs": 10},
    }


@pytest.fixture
def mock_resulter(mock_cfgs):
    return Resulter(mock_cfgs)


def create_mock_netcdf(file_path, data):
    ds = xr.Dataset({"data": (("basin", "time"), data)})
    ds.to_netcdf(file_path)


def test_load_result(mock_resulter, tmpdir):
    # Create mock prediction and observation NetCDF files
    pred_data = np.random.rand(10, 10)
    obs_data = np.random.rand(10, 10)

    pred_file = os.path.join(tmpdir, "epoch10flow_pred.nc")
    obs_file = os.path.join(tmpdir, "epoch10flow_obs.nc")

    create_mock_netcdf(pred_file, pred_data)
    create_mock_netcdf(obs_file, obs_data)

    # Load the results using the method
    pred, obs = mock_resulter.load_result()

    # Check if the loaded data matches the mock data
    np.testing.assert_array_equal(pred["data"].values, pred_data)
    np.testing.assert_array_equal(obs["data"].values, obs_data)


def test_eval_result(mock_resulter):
    # Mock configuration for eval_result
    mock_resulter.cfgs["data_cfgs"]["target_cols"] = [
        "streamflow",
        "total_evaporation_hourly",
    ]
    mock_resulter.cfgs["evaluation_cfgs"]["metrics"] = ["RMSE", "NSE"]
    mock_resulter.cfgs["evaluation_cfgs"]["fill_nan"] = ["no", "no"]
    mock_resulter.cfgs["evaluation_cfgs"]["explainer"] = "none"

    # Create mock prediction and observation xarray datasets
    preds_data = {
        "streamflow": (("time", "basin"), np.random.rand(10, 10)),
        "total_evaporation_hourly": (("time", "basin"), np.random.rand(10, 10)),
    }
    obss_data = {
        "streamflow": (("time", "basin"), np.random.rand(10, 10)),
        "total_evaporation_hourly": (("time", "basin"), np.random.rand(10, 10)),
    }
    preds_xr = xr.Dataset(preds_data)
    obss_xr = xr.Dataset(obss_data)

    # Mock the calculate_and_record_metrics function
    def mock_calculate_and_record_metrics(obs, pred, metrics, col, fill_nan, eval_log):
        eval_log[col] = {metric: np.random.rand() for metric in metrics}
        return eval_log

    mock_resulter.calculate_and_record_metrics = mock_calculate_and_record_metrics

    # Evaluate the results
    eval_log = mock_resulter.eval_result(preds_xr, obss_xr)

    # Check if the evaluation log contains the expected metrics for each variable
    for col in mock_resulter.cfgs["data_cfgs"]["target_cols"]:
        for metric in mock_resulter.cfgs["evaluation_cfgs"]["metrics"]:
            assert f"{metric} of {col}" in eval_log
            assert isinstance(eval_log[f"{metric} of {col}"], list)


"""
Author: Wenyu Ouyang
Date: 2023-09-18 14:34:53
LastEditTime: 2024-04-11 09:06:09
LastEditors: Wenyu Ouyang
Description: A simple evaluate model test
FilePath: \torchhydro\tests\test_evaluate_model.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import pytest

from hydroutils.hydro_plot import plot_ts
from torchhydro.configs.config import update_cfg
from torchhydro.trainers.deep_hydro import DeepHydro
from torchhydro.trainers.trainer import set_random_seed


@pytest.fixture()
def camelslstm_config(args, config_data):
    args.model_loader = {"load_way": "latest"}
    args.train_mode = False
    update_cfg(config_data, args)
    return config_data


@pytest.fixture()
def camelsmtllstm_config(mtl_args, config_data):
    mtl_args.model_loader = {"load_way": "latest"}
    mtl_args.train_mode = False
    update_cfg(config_data, mtl_args)
    return config_data


@pytest.fixture
def _config(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    "_config", ["camelslstm_config", "camelsmtllstm_config"], indirect=True
)
def test_evaluate_model(_config):
    random_seed = _config["training_cfgs"]["random_seed"]
    set_random_seed(random_seed)
    model = DeepHydro(_config)
    preds_xr, obss_xr = model.model_evaluate()
    plot_ts(
        [preds_xr["time"].values, obss_xr["time"].values],
        [
            preds_xr["streamflow"].sel(basin="01013500").values,
            obss_xr["streamflow"].sel(basin="01013500").values,
        ],
        leg_lst=["pred", "obs"],
        fig_size=(6, 4),
        xlabel="Date",
        ylabel="Streamflow",
    )



"""
Author: Wenyu Ouyang
Date: 2024-04-08 18:17:44
LastEditTime: 2024-07-10 19:41:48
LastEditors: Wenyu Ouyang
Description: normalize the data
FilePath: /torchhydro/torchhydro/datasets/data_scalers.py
Copyright (c) 2024-2024 Wenyu Ouyang. All rights reserved.
"""

import copy
import json
import os
import pickle as pkl
import shutil
import pint_xarray  # noqa: F401
import xarray as xr
import numpy as np
from shutil import SameFileError
from sklearn.preprocessing import (
    StandardScaler,
    RobustScaler,
    MinMaxScaler,
    MaxAbsScaler,
)

from hydroutils.hydro_stat import (
    cal_stat_prcp_norm,
    cal_stat_gamma,
    cal_stat,
    cal_4_stat_inds,
)

from torchhydro.datasets.data_utils import (
    _trans_norm,
    _prcp_norm,
    wrap_t_s_dict,
    unify_streamflow_unit,
)

SCALER_DICT = {
    "StandardScaler": StandardScaler,
    "RobustScaler": RobustScaler,
    "MinMaxScaler": MinMaxScaler,
    "MaxAbsScaler": MaxAbsScaler,
}


class ScalerHub(object):
    """
    A class for Scaler
    """

    def __init__(
        self,
        target_vars: np.array,
        relevant_vars: np.array,
        constant_vars: np.array = None,
        data_cfgs: dict = None,
        is_tra_val_te: str = None,
        data_source: object = None,
        **kwargs,
    ):
        """
        Perform normalization

        Parameters
        ----------
        target_vars
            output variables
        relevant_vars
            dynamic input variables
        constant_vars
            static input variables
        other_vars
            other required variables
        data_cfgs
            configs for reading data
        is_tra_val_te
            train, valid or test
        kwargs
            other optional parameters for ScalerHub
        """
        self.data_cfgs = data_cfgs
        norm_keys = ["target_vars", "relevant_vars", "constant_vars"]
        norm_dict = {}
        scaler_type = data_cfgs["scaler"]
        if scaler_type == "DapengScaler":
            gamma_norm_cols = data_cfgs["scaler_params"]["gamma_norm_cols"]
            prcp_norm_cols = data_cfgs["scaler_params"]["prcp_norm_cols"]
            pbm_norm = data_cfgs["scaler_params"]["pbm_norm"]
            scaler = DapengScaler(
                target_vars,
                relevant_vars,
                constant_vars,
                data_cfgs,
                is_tra_val_te,
                prcp_norm_cols=prcp_norm_cols,
                gamma_norm_cols=gamma_norm_cols,
                pbm_norm=pbm_norm,
                data_source=data_source,
            )
            x, y, c = scaler.load_data()
            self.target_scaler = scaler

        elif scaler_type in SCALER_DICT.keys():
            # TODO: not fully tested, espacially for pbm models
            all_vars = [target_vars, relevant_vars, constant_vars]
            for i in range(len(all_vars)):
                data_tmp = all_vars[i]
                scaler = SCALER_DICT[scaler_type]()
                if data_tmp.ndim == 3:
                    # for forcings and outputs
                    num_instances, num_time_steps, num_features = data_tmp.transpose(
                        "basin", "time", "variable"
                    ).shape
                    data_tmp = data_tmp.to_numpy().reshape(-1, num_features)
                    save_file = os.path.join(
                        data_cfgs["test_path"], f"{norm_keys[i]}_scaler.pkl"
                    )
                    if is_tra_val_te == "train" and data_cfgs["stat_dict_file"] is None:
                        data_norm = scaler.fit_transform(data_tmp)
                        # Save scaler in test_path for valid/test
                        with open(save_file, "wb") as outfile:
                            pkl.dump(scaler, outfile)
                    else:
                        if data_cfgs["stat_dict_file"] is not None:
                            shutil.copy(data_cfgs["stat_dict_file"], save_file)
                        if not os.path.isfile(save_file):
                            raise FileNotFoundError(
                                "Please genereate xx_scaler.pkl file"
                            )
                        with open(save_file, "rb") as infile:
                            scaler = pkl.load(infile)
                            data_norm = scaler.transform(data_tmp)
                    data_norm = data_norm.reshape(
                        num_instances, num_time_steps, num_features
                    )
                else:
                    # for attributes
                    save_file = os.path.join(
                        data_cfgs["test_path"], f"{norm_keys[i]}_scaler.pkl"
                    )
                    if is_tra_val_te == "train" and data_cfgs["stat_dict_file"] is None:
                        data_norm = scaler.fit_transform(data_tmp)
                        # Save scaler in test_path for valid/test
                        with open(save_file, "wb") as outfile:
                            pkl.dump(scaler, outfile)
                    else:
                        if data_cfgs["stat_dict_file"] is not None:
                            shutil.copy(data_cfgs["stat_dict_file"], save_file)
                        assert os.path.isfile(save_file)
                        with open(save_file, "rb") as infile:
                            scaler = pkl.load(infile)
                            data_norm = scaler.transform(data_tmp)
                norm_dict[norm_keys[i]] = data_norm
                if i == 0:
                    self.target_scaler = scaler
            x_ = norm_dict["relevant_vars"]
            y_ = norm_dict["target_vars"]
            c_ = norm_dict["constant_vars"]
            # TODO: need more test for real data
            x = xr.DataArray(
                x_,
                coords={
                    "basin": target_vars.coords["basin"],
                    "time": target_vars.coords["time"],
                    "variable": target_vars.coords["variable"],
                },
                dims=["basin", "time", "variable"],
            )
            y = xr.DataArray(
                y_,
                coords={
                    "basin": relevant_vars.coords["basin"],
                    "time": relevant_vars.coords["time"],
                    "variable": relevant_vars.coords["variable"],
                },
                dims=["basin", "time", "variable"],
            )
            c = xr.DataArray(
                c_,
                coords={
                    "basin": constant_vars.coords["basin"],
                    "variable": constant_vars.coords["variable"],
                },
                dims=["basin", "variable"],
            )
        else:
            raise NotImplementedError(
                "We don't provide this Scaler now!!! Please choose another one: DapengScaler or key in SCALER_DICT"
            )
        print("Finish Normalization\n")
        self.x = x
        self.y = y
        self.c = c


class DapengScaler(object):
    def __init__(
        self,
        target_vars: np.array,
        relevant_vars: np.array,
        constant_vars: np.array,
        data_cfgs: dict,
        is_tra_val_te: str,
        other_vars: dict = None,
        prcp_norm_cols=None,
        gamma_norm_cols=None,
        pbm_norm=False,
        data_source: object = None,
    ):
        """
        The normalization and denormalization methods from Dapeng's 1st WRR paper.
        Some use StandardScaler, and some use special norm methods

        Parameters
        ----------
        target_vars
            output variables
        relevant_vars
            input dynamic variables
        constant_vars
            input static variables
        data_cfgs
            data parameter config in data source
        is_tra_val_te
            train/valid/test
        other_vars
            if more input are needed, list them in other_vars
        prcp_norm_cols
            data items which use _prcp_norm method to normalize
        gamma_norm_cols
            data items which use log(\sqrt(x)+.1) method to normalize
        pbm_norm
            if true, use pbm_norm method to normalize; the output of pbms is not normalized data, so its inverse is different.
        """
        if prcp_norm_cols is None:
            prcp_norm_cols = [
                "streamflow",
            ]
        if gamma_norm_cols is None:
            gamma_norm_cols = [
                "gpm_tp",
                "sta_tp",
                "total_precipitation_hourly",
                "temperature_2m",
                "dewpoint_temperature_2m",
                "surface_net_solar_radiation",
                "sm_surface",
                "sm_rootzone",
            ]
        self.data_target = target_vars
        self.data_forcing = relevant_vars
        self.data_attr = constant_vars
        self.data_cfgs = data_cfgs
        self.t_s_dict = wrap_t_s_dict(data_cfgs, is_tra_val_te)
        self.data_other = other_vars
        self.prcp_norm_cols = prcp_norm_cols
        self.gamma_norm_cols = gamma_norm_cols
        # both prcp_norm_cols and gamma_norm_cols use log(\sqrt(x)+.1) method to normalize
        self.log_norm_cols = gamma_norm_cols + prcp_norm_cols
        self.pbm_norm = pbm_norm
        self.data_source = data_source
        # save stat_dict of training period in test_path for valid/test
        stat_file = os.path.join(data_cfgs["test_path"], "dapengscaler_stat.json")
        # for testing sometimes such as pub cases, we need stat_dict_file from trained dataset
        if is_tra_val_te == "train" and data_cfgs["stat_dict_file"] is None:
            self.stat_dict = self.cal_stat_all()
            with open(stat_file, "w") as fp:
                json.dump(self.stat_dict, fp)
        else:
            # for valid/test, we need to load stat_dict from train
            if data_cfgs["stat_dict_file"] is not None:
                # we used a assigned stat file, typically for PUB exps
                # shutil.copy(data_cfgs["stat_dict_file"], stat_file)
                try:
                    shutil.copy(data_cfgs["stat_dict_file"], stat_file)
                except SameFileError:
                    print(
                        f"The source file and the target file are the same: {data_cfgs['stat_dict_file']}, skipping the copy operation."
                    )
                except Exception as e:
                    print(f"Error: {e}")
            assert os.path.isfile(stat_file)
            with open(stat_file, "r") as fp:
                self.stat_dict = json.load(fp)

    @property
    def mean_prcp(self):
        return (
            self.data_source.read_mean_prcp(self.t_s_dict["sites_id"])
            .to_array()
            .to_numpy()
            .T  # TODO: check why T is needed
        )

    def inverse_transform(self, target_values):
        """
        Denormalization for output variables

        Parameters
        ----------
        target_values
            output variables

        Returns
        -------
        np.array
            denormalized predictions
        """
        stat_dict = self.stat_dict
        target_cols = self.data_cfgs["target_cols"]
        if self.pbm_norm:
            # for pbm's output, its unit is mm/day, so we don't need to recover its unit
            pred = target_values
        else:
            pred = _trans_norm(
                target_values,
                target_cols,
                stat_dict,
                log_norm_cols=self.log_norm_cols,
                to_norm=False,
            )
            for i in range(len(self.data_cfgs["target_cols"])):
                var = self.data_cfgs["target_cols"][i]
                pred.loc[dict(variable=var)] = _prcp_norm(
                    pred.sel(variable=var).to_numpy(),
                    self.mean_prcp,
                    to_norm=False,
                )
        # add attrs for units
        pred.attrs.update(self.data_target.attrs)
        return pred.to_dataset(dim="variable")

    def cal_stat_all(self):
        """
        Calculate statistics of outputs(streamflow etc), and inputs(forcing and attributes)

        Returns
        -------
        dict
            a dict with statistic values
        """
        # streamflow, et, ssm, etc
        target_cols = self.data_cfgs["target_cols"]
        stat_dict = {}
        for i in range(len(target_cols)):
            var = target_cols[i]
            if var in self.prcp_norm_cols:
                stat_dict[var] = cal_stat_prcp_norm(
                    self.data_target.sel(variable=var).to_numpy(),
                    self.mean_prcp,
                )
            elif var in self.gamma_norm_cols:
                stat_dict[var] = cal_stat_gamma(
                    self.data_target.sel(variable=var).to_numpy()
                )
            else:
                stat_dict[var] = cal_stat(self.data_target.sel(variable=var).to_numpy())

        # forcing
        forcing_lst = self.data_cfgs["relevant_cols"]
        x = self.data_forcing
        for k in range(len(forcing_lst)):
            var = forcing_lst[k]
            if var in self.gamma_norm_cols:
                stat_dict[var] = cal_stat_gamma(x.sel(variable=var).to_numpy())
            else:
                stat_dict[var] = cal_stat(x.sel(variable=var).to_numpy())

        # const attribute
        attr_data = self.data_attr
        attr_lst = self.data_cfgs["constant_cols"]
        for k in range(len(attr_lst)):
            var = attr_lst[k]
            stat_dict[var] = cal_stat(attr_data.sel(variable=var).to_numpy())

        return stat_dict

    def get_data_obs(self, to_norm: bool = True) -> np.array:
        """
        Get observation values

        Parameters
        ----------
        rm_nan
            if true, fill NaN value with 0
        to_norm
            if true, perform normalization

        Returns
        -------
        np.array
            the output value for modeling
        """
        stat_dict = self.stat_dict
        data = self.data_target
        out = xr.full_like(data, np.nan)
        # if we don't set a copy() here, the attrs of data will be changed, which is not our wish
        out.attrs = copy.deepcopy(data.attrs)
        target_cols = self.data_cfgs["target_cols"]
        for i in range(len(target_cols)):
            var = target_cols[i]
            if var in self.prcp_norm_cols:
                out.loc[dict(variable=var)] = _prcp_norm(
                    data.sel(variable=var).to_numpy(),
                    self.mean_prcp,
                    to_norm=True,
                )
            else:
                out.loc[dict(variable=var)] = data.sel(variable=var).to_numpy()
            out.attrs["units"][var] = "dimensionless"
        out = _trans_norm(
            out,
            target_cols,
            stat_dict,
            log_norm_cols=self.log_norm_cols,
            to_norm=to_norm,
        )
        return out

    def get_data_ts(self, to_norm=True) -> np.array:
        """
        Get dynamic input data

        Parameters
        ----------
        rm_nan
            if true, fill NaN value with 0
        to_norm
            if true, perform normalization

        Returns
        -------
        np.array
            the dynamic inputs for modeling
        """
        stat_dict = self.stat_dict
        var_lst = self.data_cfgs["relevant_cols"]
        data = self.data_forcing
        data = _trans_norm(
            data, var_lst, stat_dict, log_norm_cols=self.log_norm_cols, to_norm=to_norm
        )
        return data

    def get_data_const(self, to_norm=True) -> np.array:
        """
        Attr data and normalization

        Parameters
        ----------
        rm_nan
            if true, fill NaN value with 0
        to_norm
            if true, perform normalization

        Returns
        -------
        np.array
            the static inputs for modeling
        """
        stat_dict = self.stat_dict
        var_lst = self.data_cfgs["constant_cols"]
        data = self.data_attr
        data = _trans_norm(data, var_lst, stat_dict, to_norm=to_norm)
        return data

    def load_data(self):
        """
        Read data and perform normalization for DL models

        Returns
        -------
        tuple
            x: 3-d  gages_num*time_num*var_num
            y: 3-d  gages_num*time_num*1
            c: 2-d  gages_num*var_num
        """
        x = self.get_data_ts()
        y = self.get_data_obs()
        c = self.get_data_const()
        return x, y, c



import glob

import hydrodatasource.configs.config as hdscc
import numpy as np
import pandas as pd
import xarray


def test_merge_minio_streamflow():
    '''
    basin_671_gdf = gpd.read_file("/ftproot/671_shp/671-hyd_na_dir_30s.shp")
    basin_671_gdf['STAID'] = basin_671_gdf['STAID'].apply(lambda x: x.split('-')[0].zfill(8))
    flow_files = hdscc.FS.glob(
        's3://stations-origin/zq_stations/hour_data/1h/usgs_datas_462_basins_after_2019/**')[1:]
    '''
    flow_files = glob.glob("/ftproot/camels_hourly/data/usgs_streamflow_csv/*", recursive=True)
    result_df = pd.DataFrame()
    for file in flow_files:
        '''
        hru_id = file.split('.')[0].split('_')[-1]
        if hru_id not in ['03049000', '03592718']:
            flow_df = pd.read_csv(hdscc.FS.open(file), engine='c', parse_dates=['datetime'])
        else:
            continue
        flow_df['site_no'] = flow_df['site_no'].astype(str).apply(lambda x: x.zfill(8))
        flow_df['datetime'] = flow_df['datetime'].dt.tz_localize(None)
        if '00060_discontinued' in flow_df.columns:
            flow_df = flow_df.rename(columns={'datetime': 'time_start', 'site_no': 'basin_id', '00060_discontinued': 'streamflow'})
        # 这种站只有水位，没有流量
        elif ('00060' not in flow_df.columns) & ('00060_discontinued' not in flow_df.columns):
            continue
        else:
            flow_df = flow_df.rename(columns={'datetime': 'time_start', 'site_no': 'basin_id', '00060': 'streamflow'})
        flow_df['streamflow'] = flow_df['streamflow'].astype(float)
        flow_df = flow_df[['basin_id', 'time_start', 'streamflow']]
        flow_df = flow_df.set_index('time_start').drop(columns=['basin_id'])
        flow_df = flow_df.resample('h').mean()
        hru_id_series = pd.DataFrame({'basin_id': np.repeat(hru_id, len(flow_df))})
        flow_df = pd.concat([flow_df.reset_index(), hru_id_series], axis=1)
        hru_area = basin_671_gdf['AREA'][basin_671_gdf['STAID'] == hru_id].to_list()[0]
        flow_df['streamflow'] = flow_df['streamflow'].apply(lambda x: x*3.6/(hru_area*35.31))
        '''
        hru_id = file.split('-')[0]
        # camels_hourly_csv_path = f'/ftproot/camels_hourly/data/usgs_streamflow_csv/{hru_id}-usgs-hourly.csv'
        camels_hourly_df = pd.read_csv(file, engine='c', parse_dates=['date'])
        camels_hourly_df = camels_hourly_df.rename(columns={'date': 'time_start', 'QObs(mm/h)': 'streamflow'})
        camels_hourly_df = camels_hourly_df[(camels_hourly_df['time_start'] >= '2000-01-01 00:00:00') &
                                            (camels_hourly_df['time_start'] <= '2014-12-31 23:00:00')].reset_index()
        camels_part_series = pd.DataFrame({'basin_id': np.repeat(hru_id, len(camels_hourly_df))})
        camels_part_df = pd.concat([camels_hourly_df, camels_part_series], axis=1)
        camels_part_df = camels_part_df[['basin_id', 'time_start', 'streamflow']]
        # concat_df = pd.concat([camels_part_df, flow_df], axis=0)
        result_df = pd.concat([result_df, camels_part_df], axis=0)
    result_df['time_start'] = pd.to_datetime(result_df['time_start'])
    result_df = result_df.set_index(['time_start', 'basin_id'])
    result_df = result_df[~result_df.index.duplicated()]
    result_ds = xarray.Dataset.from_dataframe(result_df)
    result_ds.to_netcdf('516_basins_2000_2014.nc')


def test_merge_chn_streamflow():
    chn_shapes = [file for file in hdscc.FS.glob('s3://basins-origin/basin_shapefiles/**') if 'CHN' in file]
    chn_stcds = [shape_id.split('.')[0].split('_')[-1] for shape_id in chn_shapes]
    chn_zq_files = [f's3://stations-origin/zq_stations/hour_data/1h/zq_CHN_songliao_{stcd}.csv' for stcd in chn_stcds]
    chn_zz_files = [f's3://stations-origin/zz_stations/hour_data/1h/zz_CHN_songliao_{stcd}.csv' for stcd in chn_stcds]
    chn_files = chn_zz_files + chn_zq_files
    streamflow_df = pd.DataFrame()
    for zq_file in chn_files:
        if hdscc.FS.exists(zq_file):
            # np.count_nonzero(np.array([hdscc.FS.exists(zq_file) for zq_file in chn_files]) == True) = 23
            zq_df = pd.read_csv(zq_file, engine='c', parse_dates=['TM'], storage_options=hdscc.MINIO_PARAM)
            zq_df['TM'] = zq_df['TM'] - np.timedelta64(8, 'h')
            zq_df_split = zq_df[(zq_df['TM'] >= '2000-01-01 00:00:00') & (zq_df['TM'] <= '2014-12-31 23:00:00')]
            zq_df_split = zq_df_split.rename(columns={'STCD': 'basin_id', 'TM': 'time_start', 'Q': 'streamflow'})
            zq_df_split = zq_df_split[['basin_id', 'time_start', 'streamflow']]
            streamflow_df = pd.concat([streamflow_df, zq_df_split])
    streamflow_df['time_start'] = pd.to_datetime(streamflow_df['time_start'])
    streamflow_df['basin_id'] = streamflow_df['basin_id'].astype(int).astype(str)
    streamflow_df = streamflow_df.set_index(['time_start', 'basin_id'])
    streamflow_df = streamflow_df[~streamflow_df.index.duplicated()]
    result_ds = xarray.Dataset.from_dataframe(streamflow_df)
    result_ds.to_netcdf('23_basins_2000_2014.nc')




"""
Author: Wenyu Ouyang
Date: 2021-12-31 11:08:29
LastEditTime: 2024-05-17 11:29:13
LastEditors: Wenyu Ouyang
Description: Loss functions
FilePath: \torchhydro\torchhydro\models\crits.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

from typing import Union
import torch
from torch import distributions as tdist, Tensor
from torchhydro.models.model_utils import get_the_device


def deal_gap_data(output, target, data_gap, device):
    """
    How to handle with gap data

    When there are NaN values in observation, we will perform a "reduce" operation on prediction.
    For example, pred = [0,1,2,3,4], obs=[5, nan, nan, 6, nan]; the "reduce" is sum;
    then, pred_sum = [0+1+2, 3+4], obs_sum=[5,6], loss = loss_func(pred_sum, obs_sum).
    Notice: when "sum", actually final index is not chosen,
    because the whole observation may be [5, nan, nan, 6, nan, nan, 7, nan, nan], 6 means sum of three elements.
    Just as the rho is 5, the final one is not chosen

    Parameters
    ----------
    output
        model output for k-th variable
    target
        target for k-th variable
    data_gap
        data_gap=1: reduce is sum
        data_gap=2: reduce is mean
    device
        where to save the data

    Returns
    -------
    tuple[tensor, tensor]
        output and target after dealing with gap
    """
    # all members in a batch has different NaN-gap, so we need a loop
    seg_p_lst = []
    seg_t_lst = []
    for j in range(target.shape[1]):
        non_nan_idx = torch.nonzero(
            ~torch.isnan(target[:, j]), as_tuple=False
        ).squeeze()
        if len(non_nan_idx) < 1:
            raise ArithmeticError("All NaN elements, please check your data")

        # 使用 cumsum 生成 scatter_index
        is_not_nan = ~torch.isnan(target[:, j])
        cumsum_is_not_nan = torch.cumsum(is_not_nan.to(torch.int), dim=0)
        first_non_nan = non_nan_idx[0]
        scatter_index = torch.full_like(
            target[:, j], fill_value=-1, dtype=torch.long
        )  # 将所有值初始化为 -1
        scatter_index[first_non_nan:] = cumsum_is_not_nan[first_non_nan:] - 1
        scatter_index = scatter_index.to(device=device)

        # 创建掩码，只保留有效的索引
        valid_mask = scatter_index >= 0

        if data_gap == 1:
            seg = torch.zeros(
                len(non_nan_idx), device=device, dtype=output.dtype
            ).scatter_add_(0, scatter_index[valid_mask], output[valid_mask, j])
            # for sum, better exclude final non-nan value as it didn't include all necessary periods
            seg_p_lst.append(seg[:-1])
            seg_t_lst.append(target[non_nan_idx[:-1], j])

        elif data_gap == 2:
            counts = torch.zeros(
                len(non_nan_idx), device=device, dtype=output.dtype
            ).scatter_add_(
                0,
                scatter_index[valid_mask],
                torch.ones_like(output[valid_mask, j], dtype=output.dtype),
            )
            seg = torch.zeros(
                len(non_nan_idx), device=device, dtype=output.dtype
            ).scatter_add_(0, scatter_index[valid_mask], output[valid_mask, j])
            seg = seg / counts.clamp(min=1)
            # for mean, we can include all periods
            seg_p_lst.append(seg)
            seg_t_lst.append(target[non_nan_idx, j])
        else:
            raise NotImplementedError(
                "We have not provided this reduce way now!! Please choose 1 or 2!!"
            )

    p = torch.cat(seg_p_lst)
    t = torch.cat(seg_t_lst)
    return p, t


class SigmaLoss(torch.nn.Module):
    def __init__(self, prior="gauss"):
        super(SigmaLoss, self).__init__()
        self.reduction = "elementwise_mean"
        self.prior = None if prior == "" else prior.split("+")

    def forward(self, output, target):
        ny = target.shape[-1]
        lossMean = 0
        for k in range(ny):
            p0 = output[:, :, k * 2]
            s0 = output[:, :, k * 2 + 1]
            t0 = target[:, :, k]
            mask = t0 == t0
            p = p0[mask]
            s = s0[mask]
            t = t0[mask]
            if self.prior[0] == "gauss":
                loss = torch.exp(-s).mul((p - t) ** 2) / 2 + s / 2
            elif self.prior[0] == "invGamma":
                c1 = float(self.prior[1])
                c2 = float(self.prior[2])
                nt = p.shape[0]
                loss = (
                    torch.exp(-s).mul((p - t) ** 2 + c2 / nt) / 2
                    + (1 / 2 + c1 / nt) * s
                )
            lossMean = lossMean + torch.mean(loss)
        return lossMean


class NSELoss(torch.nn.Module):
    # Same as Fredrick 2019
    def __init__(self):
        super(NSELoss, self).__init__()

    def forward(self, output, target):
        Ngage = target.shape[1]
        losssum = 0
        nsample = 0
        for ii in range(Ngage):
            t0 = target[:, ii, 0]
            mask = t0 == t0
            if len(mask[mask]) > 0:
                p0 = output[:, ii, 0]
                p = p0[mask]
                t = t0[mask]
                tmean = t.mean()
                SST = torch.sum((t - tmean) ** 2)
                SSRes = torch.sum((t - p) ** 2)
                temp = SSRes / ((torch.sqrt(SST) + 0.1) ** 2)
                # original NSE
                # temp = SSRes / SST
                losssum = losssum + temp
                nsample = nsample + 1
        return losssum / nsample


class MASELoss(torch.nn.Module):
    def __init__(self, baseline_method):
        """
        This implements the MASE loss function (e.g. MAE_MODEL/MAE_NAIEVE)
        """
        super(MASELoss, self).__init__()
        self.method_dict = {
            "mean": lambda x, y: torch.mean(x, 1).unsqueeze(1).repeat(1, y[1], 1)
        }
        self.baseline_method = self.method_dict[baseline_method]

    def forward(
        self, target: torch.Tensor, output: torch.Tensor, train_data: torch.Tensor, m=1
    ) -> torch.Tensor:
        # Ugh why can't all tensors have batch size... Fixes for modern
        if len(train_data.shape) < 3:
            train_data = train_data.unsqueeze(0)
        if m == 1 and len(target.shape) == 1:
            output = output.unsqueeze(0)
            output = output.unsqueeze(2)
            target = target.unsqueeze(0)
            target = target.unsqueeze(2)
        if len(target.shape) == 2:
            output = output.unsqueeze(0)
            target = target.unsqueeze(0)
        result_baseline = self.baseline_method(train_data, output.shape)
        MAE = torch.nn.L1Loss()
        mae2 = MAE(output, target)
        mase4 = MAE(result_baseline, target)
        # Prevent divison by zero/loss exploding
        if mase4 < 0.001:
            mase4 = 0.001
        return mae2 / mase4


class RMSELoss(torch.nn.Module):
    def __init__(self, variance_penalty=0.0):
        """
        Calculate RMSE

        using:
            target -> True y
            output -> Prediction by model
            source: https://discuss.pytorch.org/t/rmse-loss-function/16540/3

        Parameters
        ----------
        variance_penalty
            penalty for big variance; default is 0
        """
        super().__init__()
        self.mse = torch.nn.MSELoss()
        self.variance_penalty = variance_penalty

    def forward(self, output: torch.Tensor, target: torch.Tensor):
        if len(output) <= 1 or self.variance_penalty <= 0.0:
            return torch.sqrt(self.mse(target, output))
        diff = torch.sub(target, output)
        std_dev = torch.std(diff)
        var_penalty = self.variance_penalty * std_dev

        return torch.sqrt(self.mse(target, output)) + var_penalty


class MAPELoss(torch.nn.Module):
    """
    Returns MAPE using:
    target -> True y
    output -> Predtion by model
    """

    def __init__(self, variance_penalty=0.0):
        super().__init__()
        self.variance_penalty = variance_penalty

    def forward(self, output: torch.Tensor, target: torch.Tensor):
        if len(output) > 1:
            return torch.mean(
                torch.abs(torch.sub(target, output) / target)
            ) + self.variance_penalty * torch.std(torch.sub(target, output))
        else:
            return torch.mean(torch.abs(torch.sub(target, output) / target))


class MAELoss(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, output: torch.Tensor, target: torch.Tensor):
        # Create a mask to filter out NaN values
        mask = ~torch.isnan(target)

        # Apply the mask to both target and output
        target = target[mask]
        output = output[mask]

        # Calculate MAE for the non-NaN values
        return torch.mean(torch.abs(target - output))


class PenalizedMSELoss(torch.nn.Module):
    """
    Returns MSE using:
    target -> True y
    output -> Predtion by model
    source: https://discuss.pytorch.org/t/rmse-loss-function/16540/3
    """

    def __init__(self, variance_penalty=0.0):
        super().__init__()
        self.mse = torch.nn.MSELoss()
        self.variance_penalty = variance_penalty

    def forward(self, output: torch.Tensor, target: torch.Tensor):
        return self.mse(target, output) + self.variance_penalty * torch.std(
            torch.sub(target, output)
        )


class GaussianLoss(torch.nn.Module):
    def __init__(self, mu=0, sigma=0):
        """Compute the negative log likelihood of Gaussian Distribution
        From https://arxiv.org/abs/1907.00235
        """
        super(GaussianLoss, self).__init__()
        self.mu = mu
        self.sigma = sigma

    def forward(self, x: torch.Tensor):
        loss = -tdist.Normal(self.mu, self.sigma).log_prob(x)
        return torch.sum(loss) / (loss.size(0) * loss.size(1))


class QuantileLoss(torch.nn.Module):
    """From https://medium.com/the-artificial-impostor/quantile-regression-part-2-6fdbc26b2629"""

    def __init__(self, quantiles):
        super().__init__()
        self.quantiles = quantiles

    def forward(self, preds, target):
        assert not target.requires_grad
        assert preds.size(0) == target.size(0)
        losses = []
        for i, q in enumerate(self.quantiles):
            mask = ~torch.isnan(target[:, :, i])
            errors = target[:, :, i][mask] - preds[:, :, i][mask]
            losses.append(torch.max((q - 1) * errors, q * errors))
        return torch.mean(torch.cat(losses, dim=0), dim=0)


class NegativeLogLikelihood(torch.nn.Module):
    """
    target -> True y
    output -> predicted distribution
    """

    def __init__(self):
        super().__init__()

    def forward(self, output: torch.distributions, target: torch.Tensor):
        """
        calculates NegativeLogLikelihood
        """
        return -output.log_prob(target).sum()


def l1_regularizer(model, lambda_l1=0.01):
    """
    source: https://stackoverflow.com/questions/58172188/how-to-add-l1-regularization-to-pytorch-nn-model
    """
    lossl1 = 0
    for model_param_name, model_param_value in model.named_parameters():
        if model_param_name.endswith("weight"):
            lossl1 += lambda_l1 * model_param_value.abs().sum()
        return lossl1


def orth_regularizer(model, lambda_orth=0.01):
    """
    source: https://stackoverflow.com/questions/58172188/how-to-add-l1-regularization-to-pytorch-nn-model
    """
    lossorth = 0
    for model_param_name, model_param_value in model.named_parameters():
        if model_param_name.endswith("weight"):
            param_flat = model_param_value.view(model_param_value.shape[0], -1)
            sym = torch.mm(param_flat, torch.t(param_flat))
            sym -= torch.eye(param_flat.shape[0])
            lossorth += lambda_orth * sym.sum()

        return lossorth


class RmseLoss(torch.nn.Module):
    def __init__(self):
        """
        RMSE loss which could ignore NaN values

        Now we only support 3-d tensor and 1-d tensor
        """
        super(RmseLoss, self).__init__()

    def forward(self, output, target):
        if target.dim() == 1:
            mask = target == target
            p = output[mask]
            t = target[mask]
            return torch.sqrt(((p - t) ** 2).mean())
        ny = target.shape[2]
        loss = 0
        for k in range(ny):
            p0 = output[:, :, k]
            t0 = target[:, :, k]
            mask = t0 == t0
            p = p0[mask]
            p = torch.where(torch.isnan(p), torch.full_like(p, 0), p)
            t = t0[mask]
            t = torch.where(torch.isnan(t), torch.full_like(t, 0), t)
            temp = torch.sqrt(((p - t) ** 2).mean())
            loss = loss + temp
        return loss


class MultiOutLoss(torch.nn.Module):
    def __init__(
        self,
        loss_funcs: Union[torch.nn.Module, list],
        data_gap: list = None,
        device: list = None,
        limit_part: list = None,
        item_weight: list = None,
    ):
        """
        Loss function for multiple output

        Parameters
        ----------
        loss_funcs
            The loss functions for each output
        data_gap
            It belongs to the feature dim.
            If 1, then the corresponding value is uniformly-spaced with NaN values filling the gap;
            in addition, the first non-nan value means the aggregated value of the following interval,
            for example, in [5, nan, nan, nan], 5 means all four data's sum, although the next 3 values are nan
            hence the calculation is a little different;
            if 2, the first non-nan value means the average value of the following interval,
            for example, in [5, nan, nan, nan], 5 means all four data's mean value;
            default is [0, 2]
        device
            the number of device: -1 -> "cpu" or "cuda:x" (x is 0, 1 or ...)
        limit_part
            when transfer learning, we may ignore some part;
            the default is None, which means no ignorance;
            other choices are list, such as [0], [0, 1] or [1,2,..];
            0 means the first variable;
            tensor is [seq, time, var] or [time, seq, var]
        item_weight
            use different weight for each item's loss;
            for example, the default values [0.5, 0.5] means 0.5 * loss1 + 0.5 * loss2
        """
        if data_gap is None:
            data_gap = [0, 2]
        if device is None:
            device = [0]
        if item_weight is None:
            item_weight = [0.5, 0.5]
        super(MultiOutLoss, self).__init__()
        self.loss_funcs = loss_funcs
        self.data_gap = data_gap
        self.device = get_the_device(device)
        self.limit_part = limit_part
        self.item_weight = item_weight

    def forward(self, output: Tensor, target: Tensor):
        """
        Calculate the sum of losses for different variables

        When there are NaN values in observation, we will perform a "reduce" operation on prediction.
        For example, pred = [0,1,2,3,4], obs=[5, nan, nan, 6, nan]; the "reduce" is sum;
        then, pred_sum = [0+1+2, 3+4], obs_sum=[5,6], loss = loss_func(pred_sum, obs_sum).
        Notice: when "sum", actually final index is not chosen,
        because the whole observation may be [5, nan, nan, 6, nan, nan, 7, nan, nan], 6 means sum of three elements.
        Just as the rho is 5, the final one is not chosen


        Parameters
        ----------
        output
            the prediction tensor; 3-dims are time sequence, batch and feature, respectively
        target
            the observation tensor

        Returns
        -------
        Tensor
            Whole loss
        """
        n_out = target.shape[-1]
        loss = 0
        for k in range(n_out):
            if self.limit_part is not None and k in self.limit_part:
                continue
            p0 = output[:, :, k]
            t0 = target[:, :, k]
            mask = t0 == t0
            p = p0[mask]
            t = t0[mask]
            if self.data_gap[k] > 0:
                p, t = deal_gap_data(p0, t0, self.data_gap[k], self.device)
            if type(self.loss_funcs) is list:
                temp = self.item_weight[k] * self.loss_funcs[k](p, t)
            else:
                temp = self.item_weight[k] * self.loss_funcs(p, t)
            # sum of all k-th loss
            if torch.isnan(temp).any():
                continue
            loss = loss + temp
        return loss


# ref: https://github.com/median-research-group/LibMTL
class UncertaintyWeights(torch.nn.Module):
    r"""Uncertainty Weights (UW).

    This method is proposed in `Multi-Task Learning Using Uncertainty to Weigh Losses for Scene Geometry and Semantics (CVPR 2018) <https://openaccess.thecvf.com/content_cvpr_2018/papers/Kendall_Multi-Task_Learning_Using_CVPR_2018_paper.pdf>`_ \
    and implemented by us.

    """

    def __init__(
        self,
        loss_funcs: Union[torch.nn.Module, list],
        data_gap: list = None,
        device: list = None,
        limit_part: list = None,
    ):
        if data_gap is None:
            data_gap = [0, 2]
        if device is None:
            device = [0]
        super(UncertaintyWeights, self).__init__()
        self.loss_funcs = loss_funcs
        self.data_gap = data_gap
        self.device = get_the_device(device)
        self.limit_part = limit_part

    def forward(self, output, target, log_vars):
        """

        Parameters
        ----------
        output
        target
        log_vars
            sigma in uncertainty weighting;
            default is None, meaning we manually set weights for different target's loss;
            more info could be seen in
            https://libmtl.readthedocs.io/en/latest/docs/_autoapi/LibMTL/weighting/index.html#LibMTL.weighting.UW

        Returns
        -------
        torch.Tensor
            multi-task loss by uncertainty weighting method
        """
        n_out = target.shape[-1]
        loss = 0
        for k in range(n_out):
            precision = torch.exp(-log_vars[k])
            if self.limit_part is not None and k in self.limit_part:
                continue
            p0 = output[:, :, k]
            t0 = target[:, :, k]
            mask = t0 == t0
            p = p0[mask]
            t = t0[mask]
            if self.data_gap[k] > 0:
                p, t = deal_gap_data(p0, t0, self.data_gap[k], self.device)
            if type(self.loss_funcs) is list:
                temp = self.loss_funcs[k](p, t)
            else:
                temp = self.loss_funcs(p, t)
            loss += torch.sum(precision * temp + log_vars[k], -1)
        return loss


# ref: https://openaccess.thecvf.com/content_ECCV_2018/html/Michelle_Guo_Focus_on_the_ECCV_2018_paper.html
class DynamicTaskPrior(torch.nn.Module):
    r"""Dynamic Task Prioritization

    This method is proposed in https://openaccess.thecvf.com/content_ECCV_2018/html/Michelle_Guo_Focus_on_the_ECCV_2018_paper.html
    In contrast to UW and other curriculum learning methods, where easy tasks are prioritized above difficult tasks,
    It shows the importance of prioritizing difficult tasks first.
    It automatically prioritize more difficult tasks by adaptively adjusting the mixing weight of each task’s loss.
    Here we choose correlation as KPI. As KPI must be in [0,1], we set (corr+1)/2 as KPI
    """

    def __init__(
        self,
        loss_funcs: Union[torch.nn.Module, list],
        data_gap: list = None,
        device: list = None,
        limit_part: list = None,
        gamma=2,
        alpha=0.5,
    ):
        """

        Parameters
        ----------
        loss_funcs
        data_gap
        device
        limit_part
        gamma
            the example-level focusing parameter
        alpha
            default is 1, which means we only use the newest KPI value
        """
        if data_gap is None:
            data_gap = [0, 2]
        if device is None:
            device = [0]
        super(DynamicTaskPrior, self).__init__()
        self.loss_funcs = loss_funcs
        self.data_gap = data_gap
        self.device = get_the_device(device)
        self.limit_part = limit_part
        self.gamma = gamma
        self.alpha = alpha

    def forward(self, output, target, kpi_last=None):
        """
        Parameters
        ----------
        output
            model's prediction
        target
            observation

        kpi_last
            the KPI value of last iteration; each element for an output
            It use a moving average KPI as the weighting coefficient: KPI_i = alpha * KPI_i + (1-alpha) * KPI_{i-1}

        Returns
        -------
        torch.Tensor
            multi-task loss by Dynamic Task Prioritization method
        """
        n_out = target.shape[-1]
        loss = 0
        kpis = torch.zeros(n_out).to(self.device)
        for k in range(n_out):
            if self.limit_part is not None and k in self.limit_part:
                continue
            p0 = output[:, :, k]
            t0 = target[:, :, k]
            mask = t0 == t0
            p = p0[mask]
            t = t0[mask]
            if self.data_gap[k] > 0:
                p, t = deal_gap_data(p0, t0, self.data_gap[k], self.device)
            if type(self.loss_funcs) is list:
                temp = self.loss_funcs[k](p, t)
            else:
                temp = self.loss_funcs(p, t)
            # kpi must be in [0, 1], as corr's range is [-1, 1], just trans corr to  (corr+1)/2
            kpi = (torch.corrcoef(torch.stack([p, t], 1).T)[0, 1] + 1) / 2
            if self.alpha < 1:
                assert kpi_last is not None
                kpi = kpi * self.alpha + kpi_last[k] * (1 - self.alpha)
                # if we exclude kpi from the backward, it trans to a normal multi-task model
                # kpi = kpi.detach().clone() * self.alpha + kpi_last[k] * (1 - self.alpha)
            kpis[k] = kpi
            # focal loss
            fl = -((1 - kpi) ** self.gamma) * torch.log(kpi)
            loss += torch.sum(fl * temp, -1)
        # if kpi has grad_fn, backward will repeat. It won't work
        return loss, kpis.detach().clone()


class MultiOutWaterBalanceLoss(torch.nn.Module):
    def __init__(
        self,
        loss_funcs: Union[torch.nn.Module, list],
        data_gap: list = None,
        device: list = None,
        limit_part: list = None,
        item_weight: list = None,
        alpha=0.5,
        beta=0.0,
        wb_loss_func=None,
        means=None,
        stds=None,
    ):
        """
        Loss function for multiple output considering water balance

        loss = alpha * water_balance_loss + (1-alpha) * mtl_loss

        This loss function is only for p, q, et now
        we use the difference between p_obs_mean-q_obs_mean-et_obs_mean and p_pred_mean-q_pred_mean-et_pred_mean as water balance loss
        which is the difference between (q_obs_mean + et_obs_mean) and (q_pred_mean + et_pred_mean)

        Parameters
        ----------
        loss_funcs
            The loss functions for each output
        data_gap
            It belongs to the feature dim.
            If 1, then the corresponding value is uniformly-spaced with NaN values filling the gap;
            in addition, the first non-nan value means the aggregated value of the following interval,
            for example, in [5, nan, nan, nan], 5 means all four data's sum, although the next 3 values are nan
            hence the calculation is a little different;
            if 2, the first non-nan value means the average value of the following interval,
            for example, in [5, nan, nan, nan], 5 means all four data's mean value;
            default is [0, 2]
        device
            the number of device: -1 -> "cpu" or "cuda:x" (x is 0, 1 or ...)
        limit_part
            when transfer learning, we may ignore some part;
            the default is None, which means no ignorance;
            other choices are list, such as [0], [0, 1] or [1,2,..];
            0 means the first variable;
            tensor is [seq, time, var] or [time, seq, var]
        item_weight
            use different weight for each item's loss;
            for example, the default values [0.5, 0.5] means 0.5 * loss1 + 0.5 * loss2
        alpha
            the weight of the water-balance item's loss
        beta
            the weight of real water-balance item's loss, et_mean/p_mean + q_mean/p_mean = 1 can be a loss.
            It is not strictly correct as training batch only have about one year data, but still could be a constraint
        wb_loss_func
            the loss function for water balance item, by default it is None, which means we use function in loss_funcs
        """
        if data_gap is None:
            data_gap = [0, 2]
        if device is None:
            device = [0]
        if item_weight is None:
            item_weight = [0.5, 0.5]
        super(MultiOutWaterBalanceLoss, self).__init__()
        self.loss_funcs = loss_funcs
        self.data_gap = data_gap
        self.device = get_the_device(device)
        self.limit_part = limit_part
        self.item_weight = item_weight
        self.alpha = alpha
        self.beta = beta
        self.wb_loss_func = wb_loss_func
        self.means = means
        self.stds = stds

    def forward(self, output: Tensor, target: Tensor):
        """
        Calculate the sum of losses for different variables and water-balance loss

        When there are NaN values in observation, we will perform a "reduce" operation on prediction.
        For example, pred = [0,1,2,3,4], obs=[5, nan, nan, 6, nan]; the "reduce" is sum;
        then, pred_sum = [0+1+2, 3+4], obs_sum=[5,6], loss = loss_func(pred_sum, obs_sum).
        Notice: when "sum", actually final index is not chosen,
        because the whole observation may be [5, nan, nan, 6, nan, nan, 7, nan, nan], 6 means sum of three elements.
        Just as the rho is 5, the final one is not chosen


        Parameters
        ----------
        output
            the prediction tensor; 3-dims are time sequence, batch and feature, respectively
        target
            the observation tensor

        Returns
        -------
        Tensor
            Whole loss
        """
        n_out = target.shape[-1]
        loss = 0
        p_means = []
        t_means = []
        all_means = self.means
        all_stds = self.stds
        for k in range(n_out):
            if self.limit_part is not None and k in self.limit_part:
                continue
            p0 = output[:, :, k]
            t0 = target[:, :, k]
            # for water balance loss
            if all_means is not None:
                # denormalize for q and et
                p1 = p0 * all_stds[k] + all_means[k]
                t1 = t0 * all_stds[k] + all_means[k]
                p2 = (10**p1 - 0.1) ** 2
                t2 = (10**t1 - 0.1) ** 2
                p_mean = torch.nanmean(p2, dim=0)
                t_mean = torch.nanmean(t2, dim=0)
            else:
                p_mean = torch.nanmean(p0, dim=0)
                t_mean = torch.nanmean(t0, dim=0)
            p_means.append(p_mean)
            t_means.append(t_mean)
            # for mtl normal loss
            mask = t0 == t0
            p = p0[mask]
            t = t0[mask]
            if self.data_gap[k] > 0:
                p, t = deal_gap_data(p0, t0, self.data_gap[k], self.device)
            if type(self.loss_funcs) is list:
                temp = self.item_weight[k] * self.loss_funcs[k](p, t)
            else:
                temp = self.item_weight[k] * self.loss_funcs(p, t)
            # sum of all k-th loss
            loss = loss + temp
        # water balance loss
        p_mean_q_plus_et = torch.sum(torch.stack(p_means), dim=0)
        t_mean_q_plus_et = torch.sum(torch.stack(t_means), dim=0)
        wb_ones = torch.ones_like(t_mean_q_plus_et)
        if self.wb_loss_func is None:
            if type(self.loss_funcs) is list:
                # if wb_loss_func is None, we use the first loss function in loss_funcs
                wb_loss = self.loss_funcs[0](p_mean_q_plus_et, t_mean_q_plus_et)
                wb_1loss = self.loss_funcs[0](p_mean_q_plus_et, wb_ones)
            else:
                wb_loss = self.loss_funcs(p_mean_q_plus_et, t_mean_q_plus_et)
                wb_1loss = self.loss_funcs(p_mean_q_plus_et, wb_ones)
        else:
            wb_loss = self.wb_loss_func(p_mean_q_plus_et, t_mean_q_plus_et)
            wb_1loss = self.wb_loss_func(p_mean_q_plus_et, wb_ones)
        return (
            self.alpha * wb_loss
            + (1 - self.alpha - self.beta) * loss
            + self.beta * wb_1loss
        )



"""
Author: Wenyu Ouyang
Date: 2023-10-05 16:16:48
LastEditTime: 2024-02-14 16:12:50
LastEditors: Wenyu Ouyang
Description: A test function for transfer learning
FilePath: \torchhydro\tests\test_transfer_learning.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import os
import pytest
from hydroutils.hydro_file import get_lastest_file_in_a_dir

from torchhydro import SETTING
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate


@pytest.fixture()
def var_c_target():
    return [
        "elev_mean",
        "slope_mean",
        "area_gages2",
        "frac_forest",
        "lai_max",
        "lai_diff",
    ]


@pytest.fixture()
def var_c_source():
    return [
        "elev_mean",
        "slope_mean",
        "area_gages2",
        "frac_forest",
        "lai_max",
        "lai_diff",
        "dom_land_cover_frac",
        "dom_land_cover",
        "root_depth_50",
        "soil_depth_statsgo",
        "soil_porosity",
        "soil_conductivity",
        "max_water_content",
        "geol_1st_class",
        "geol_2nd_class",
        "geol_porostiy",
        "geol_permeability",
    ]


@pytest.fixture()
def var_t_target():
    return ["dayl", "prcp", "srad"]


@pytest.fixture()
def var_t_source():
    return ["dayl", "prcp", "srad", "tmax", "tmin", "vp"]


def test_transfer_gages_lstm_model(
    var_c_source, var_c_target, var_t_source, var_t_target
):
    weight_dir = os.path.join(
        os.getcwd(),
        "results",
        "test_camels",
        "exp1",
    )
    weight_path = get_lastest_file_in_a_dir(weight_dir)
    project_name = "test_camels/exp4"
    args = cmd(
        sub=project_name,
        source="CAMELS",
        source_path=os.path.join(
            SETTING["local_data_path"]["datasets-origin"], "camels", "camels_us"
        ),
        source_region="US",
        ctx=[0],
        model_type="TransLearn",
        model_name="KaiLSTM",
        model_hyperparam={
            "linear_size": len(var_c_target) + len(var_t_target),
            "n_input_features": len(var_c_source) + len(var_t_source),
            "n_output_features": 1,
            "n_hidden_states": 256,
        },
        opt="Adadelta",
        loss_func="RMSESum",
        batch_size=5,
        forecast_history=0,
        forecast_length=20,
        rs=1234,
        train_period=["2010-10-01", "2011-10-01"],
        test_period=["2011-10-01", "2012-10-01"],
        scaler="DapengScaler",
        sampler="KuaiSampler",
        dataset="StreamflowDataset",
        weight_path=weight_path,
        weight_path_add={
            "freeze_params": ["lstm.b_hh", "lstm.b_ih", "lstm.w_hh", "lstm.w_ih"]
        },
        continue_train=True,
        train_epoch=20,
        te=20,
        save_epoch=10,
        var_t=var_t_target,
        var_c=var_c_target,
        var_out=["streamflow"],
        gage_id=[
            "01055000",
            "01057000",
            "01170100",
        ],
    )
    cfg = default_config_file()
    update_cfg(cfg, args)
    train_and_evaluate(cfg)
    print("All processes are finished!")



"""
Author: Wenyu Ouyang
Date: 2024-05-27 13:33:08
LastEditTime: 2024-05-27 15:31:14
LastEditors: Wenyu Ouyang
Description: Unit test for datasets
FilePath: \torchhydro\tests\test_data_sets.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import pytest
import os
import numpy as np
import pandas as pd
import xarray as xr
import pickle
from sklearn.preprocessing import StandardScaler
from torchhydro.datasets.data_sets import BaseDataset
from torchhydro.datasets.data_sources import data_sources_dict


class MockDatasource:
    def __init__(self, source_cfgs):
        self.ngrid = 2
        self.nt = 366
        self.data_cfgs = source_cfgs

    def read_ts_xrdataset(self, basin_id, t_range, var_lst):
        basins = [f"{i:08d}" for i in range(1013500, 1013500 + self.ngrid)]

        # 创建时间序列
        start_date, end_date = t_range
        times = pd.date_range(start=start_date, end=end_date, freq="D")

        # 确保时间序列长度和 nt 一致
        if len(times) != self.nt:
            raise ValueError(
                "The generated time range does not match the expected length."
            )

        # 创建数据变量
        data_vars = {
            var: (["basin", "time"], np.random.rand(self.ngrid, self.nt))
            for var in var_lst
        }

        # 创建 xarray Dataset
        dataset = xr.Dataset(
            data_vars=data_vars, coords={"basin": basins, "time": times}
        )
        if "streamflow" in dataset.variables:
            dataset["streamflow"].attrs["units"] = "mm/day"

        if "prcp" in dataset.variables:
            dataset["prcp"].attrs["units"] = "mm/day"

        return dataset

    def read_attr_xrdataset(self, basin_id, var_lst, all_number=True):
        # 创建 basin ID 列表
        basins = [f"{i:08d}" for i in range(1013500, 1013500 + self.ngrid)]

        # 创建数据变量
        attr_lst = ["geol_1st_class", "geol_2nd_class"]
        data_vars = {attr: (["basin"], np.random.rand(self.ngrid)) for attr in attr_lst}

        # 添加整数类型的虚拟数据变量

        for var in attr_lst:
            data_vars[var] = (["basin"], np.random.randint(0, 10, size=self.ngrid))

        return xr.Dataset(data_vars=data_vars, coords={"basin": basins})


def test_create_lookup_table(tmp_path):
    temp_test_path = tmp_path / "test_datasets"
    os.makedirs(temp_test_path, exist_ok=True)
    data_sources_dict.update({"mockdatasource": MockDatasource})
    data_cfgs = {
        "source_cfgs": {
            "source_name": "mockdatasource",
            "source_path": str(temp_test_path),
        },
        "test_path": str(temp_test_path),
        "object_ids": [
            "01013500",
            "01013501",
        ],  # Add this line with the actual object IDs
        "t_range_train": [
            "2001-01-01",
            "2002-01-01",
        ],  # Add this line with the actual start and end dates for training.
        "t_range_test": [
            "2002-01-01",
            "2003-01-01",
        ],  # Add this line with the actual start and end dates for validation.
        "relevant_cols": [
            # List the relevant column names here.
            "prcp",
            "pet",
            # ... other relevant columns ...
        ],
        "target_cols": [
            # List the target column names here.
            "streamflow",
            "surface_sm",
            # ... other target columns ...
        ],
        "constant_cols": [
            # List the constant column names here.
            "geol_1st_class",
            "geol_2nd_class",
            # ... other constant columns ...
        ],
        "forecast_history": 7,
        "warmup_length": 14,
        "forecast_length": 1,
        "min_time_unit": "D",
        "min_time_interval": 1,
        "target_rm_nan": True,
        "relevant_rm_nan": True,
        "constant_rm_nan": True,
        "scaler": "StandardScaler",  # Add the scaler configuration here
        "stat_dict_file": None,  # Added the missing configuration
    }
    is_tra_val_te = "train"
    dataset = BaseDataset(data_cfgs, is_tra_val_te)
    lookup_table = dataset.lookup_table
    assert isinstance(lookup_table, dict)
    assert len(lookup_table) > 0
    assert all(
        isinstance(key, int) and isinstance(value, tuple)
        for key, value in lookup_table.items()
    )
    is_tra_val_te = "test"
    mock_data = np.random.rand(100, 2)  # Replace with relevant data.
    scaler = StandardScaler()
    scaler.fit(mock_data)
    scaler_file_path = tmp_path / "test_datasets_scaler.pkl"
    data_cfgs["stat_dict_file"] = str(scaler_file_path)
    with open(scaler_file_path, "wb") as file:
        pickle.dump(scaler, file)
    dataset = BaseDataset(data_cfgs, is_tra_val_te)
    lookup_table = dataset.lookup_table
    assert isinstance(lookup_table, dict)
    assert len(lookup_table) > 0
    assert all(
        isinstance(key, int) and isinstance(value, tuple)
        for key, value in lookup_table.items()
    )



import pytest
import os
from hydroutils import hydro_file

from torchhydro import SETTING
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import ensemble_train_and_evaluate


@pytest.fixture()
def var_c_target():
    return [
        "p_mean",
        "pet_mean",
        "Area",
        "geol_class_1st",
        "elev",
        "SNDPPT",
    ]


@pytest.fixture()
def var_c_source():
    return [
        "elev_mean",
        "slope_mean",
        "area_gages2",
        "frac_forest",
        "lai_max",
        "lai_diff",
        "dom_land_cover_frac",
        "dom_land_cover",
        "root_depth_50",
        "soil_depth_statsgo",
        "soil_porosity",
        "soil_conductivity",
        "max_water_content",
        "geol_1st_class",
        "geol_2nd_class",
        "geol_porostiy",
        "geol_permeability",
    ]


@pytest.fixture()
def var_t_target():
    # mainly from ERA5LAND
    return ["total_precipitation", "potential_evaporation", "temperature_2m"]


@pytest.fixture()
def var_t_source():
    return ["dayl", "prcp", "srad", "tmax", "tmin", "vp"]


@pytest.fixture()
def gage_id():
    return [
        "61561",
    ]


def test_run_lstm_cross_val(var_c_target, var_t_target, gage_id):
    config = default_config_file()
    project_name = "test_camels/expcccv61561"
    kfold = 2
    train_period = ["2018-10-01", "2021-10-01"]
    valid_period = ["2015-10-01", "2018-10-01"]
    args = cmd(
        sub=project_name,
        source="SelfMadeCAMELS",
        source_path=os.path.join(
            SETTING["local_data_path"]["datasets-interim"], "camels_cc"
        ),
        ctx=[0],
        model_name="KuaiLSTM",
        model_hyperparam={
            "n_input_features": len(var_c_target) + len(var_t_target),
            "n_output_features": 1,
            "n_hidden_states": 256,
        },
        opt="Adadelta",
        # opt_param=opt_param,
        loss_func="RMSESum",
        train_period=train_period,
        test_period=valid_period,
        batch_size=20,
        forecast_history=0,
        forecast_length=365,
        scaler="DapengScaler",
        dataset="StreamflowDataset",
        continue_train=True,
        warmup_length=0,
        train_epoch=10,
        te=10,
        var_t=var_t_target,
        var_t_type="era5land",
        var_c=var_c_target,
        var_out=["streamflow"],
        gage_id=gage_id,
        ensemble=True,
        ensemble_items={
            "kfold": kfold,
            "batch_sizes": [20, 50],
        },
    )
    update_cfg(config, args)
    ensemble_train_and_evaluate(config)
    print("All processes are finished!")


def test_run_cross_val_tlcamelsus2cc(
    var_c_source, var_c_target, var_t_source, var_t_target, gage_id
):
    weight_dir = os.path.join(
        os.getcwd(),
        "results",
        "test_camels",
        "exp1",
    )
    weight_path = hydro_file.get_lastest_file_in_a_dir(weight_dir)
    project_name = "test_camels/exptl4cccv61561"
    train_period = ["2018-10-01", "2021-10-01"]
    valid_period = ["2015-10-01", "2018-10-01"]
    args = cmd(
        sub=project_name,
        source="SelfMadeCAMELS",
        # cc means China continent
        source_path=os.path.join(
            SETTING["local_data_path"]["datasets-interim"], "camels_cc"
        ),
        ctx=[0],
        model_type="TransLearn",
        model_name="KaiLSTM",
        model_hyperparam={
            "linear_size": len(var_c_target) + len(var_t_target),
            "n_input_features": len(var_c_source) + len(var_t_source),
            "n_output_features": 1,
            "n_hidden_states": 256,
        },
        opt="Adadelta",
        loss_func="RMSESum",
        batch_size=5,
        forecast_history=0,
        forecast_length=20,
        rs=1234,
        train_period=train_period,
        test_period=valid_period,
        scaler="DapengScaler",
        # sampler="KuaiSampler",
        dataset="StreamflowDataset",
        weight_path=weight_path,
        weight_path_add={
            "freeze_params": ["lstm.b_hh", "lstm.b_ih", "lstm.w_hh", "lstm.w_ih"]
        },
        continue_train=True,
        train_epoch=10,
        te=10,
        var_t=var_t_target,
        var_c=var_c_target,
        var_out=["streamflow"],
        gage_id=gage_id,
        ensemble=True,
        ensemble_items={
            "kfold": 2,
            "batch_sizes": [20, 50],
        },
    )
    cfg = default_config_file()
    update_cfg(cfg, args)
    ensemble_train_and_evaluate(cfg)
    print("All processes are finished!")



"""
Author: Wenyu Ouyang
Date: 2024-04-17 12:55:24
LastEditTime: 2024-05-27 10:10:35
LastEditors: Wenyu Ouyang
Description:
FilePath: \torchhydro\tests\test_train_seq2seq.py
Copyright (c) 2021-2024 Wenyu Ouyang. All rights reserved.
"""

import logging
import os.path
import pathlib

import pandas as pd
import pytest
import hydrodatasource.configs.config as hdscc
import xarray as xr
import torch.multiprocessing as mp

from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.deep_hydro import train_worker
from torchhydro.trainers.trainer import train_and_evaluate

# from torchhydro.trainers.trainer import train_and_evaluate, ensemble_train_and_evaluate

logging.basicConfig(level=logging.INFO)
for logger_name in logging.root.manager.loggerDict:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

show = pd.read_csv(
    os.path.join(pathlib.Path(__file__).parent.parent, "data/basin_id(all).csv"),
    dtype={"id": str},
)
# gage_id = show["id"].values.tolist()
gage_id = ["songliao_21401550"]


@pytest.fixture()
def config():
    # 设置测试所需的项目名称和默认配置文件
    project_name = os.path.join("train_with_gpm", "ex1")
    config_data = default_config_file()

    # 填充测试所需的命令行参数
    args = cmd(
        sub=project_name,
        source_cfgs={
            "source": "HydroMean",
            "source_path": "/ftproot/basins-interim/",
        },
        ctx=[2],
        model_name="Seq2Seq",
        model_hyperparam={
            "en_input_size": 17,
            "de_input_size": 18,
            "output_size": 2,
            "hidden_size": 256,
            "forecast_length": 56,
            "prec_window": 1,
            "teacher_forcing_ratio": 0.5,
        },
        model_loader={"load_way": "best"},
        gage_id=gage_id,
        # gage_id=["21400800", "21401550", "21401300", "21401900"],
        batch_size=256,
        forecast_history=240,
        forecast_length=56,
        min_time_unit="h",
        min_time_interval=3,
        var_t=[
            "precipitationCal",
            "sm_surface",
        ],
        var_c=[
            "area",  # 面积
            "ele_mt_smn",  # 海拔(空间平均)
            "slp_dg_sav",  # 地形坡度 (空间平均)
            "sgr_dk_sav",  # 河流坡度 (平均)
            "for_pc_sse",  # 森林覆盖率
            "glc_cl_smj",  # 土地覆盖类型
            "run_mm_syr",  # 陆面径流 (流域径流的空间平均值)
            "inu_pc_slt",  # 淹没范围 (长期最大)
            "cmi_ix_syr",  # 气候湿度指数
            "aet_mm_syr",  # 实际蒸散发 (年平均)
            "snw_pc_syr",  # 雪盖范围 (年平均)
            "swc_pc_syr",  # 土壤水含量
            "gwt_cm_sav",  # 地下水位深度
            "cly_pc_sav",  # 土壤中的黏土、粉砂、砂粒含量
            "dor_pc_pva",  # 调节程度
        ],
        var_out=["streamflow", "sm_surface"],
        dataset="Seq2SeqDataset",
        sampler="HydroSampler",
        scaler="DapengScaler",
        train_epoch=100,
        save_epoch=1,
        train_period=[("2016-06-01-01", "2023-11-01-01")],
        test_period=[("2015-06-01-01", "2016-06-01-01")],
        valid_period=[("2015-06-01-01", "2016-06-01-01")],
        loss_func="MultiOutLoss",
        loss_param={
            "loss_funcs": "RMSESum",
            "data_gap": [0, 0],
            "device": [1],
            "item_weight": [0.8, 0.2],
        },
        opt="Adam",
        lr_scheduler={
            "lr": 0.0001,
            "lr_factor": 0.9,
        },
        which_first_tensor="batch",
        rolling=False,
        long_seq_pred=False,
        calc_metrics=False,
        early_stopping=True,
        # ensemble=True,
        # ensemble_items={
        #     "batch_sizes": [256, 512],
        # },
        patience=10,
        model_type="MTL",
    )

    # 更新默认配置
    update_cfg(config_data, args)

    return config_data


def test_seq2seq(config):
    # world_size = len(config["training_cfgs"]["device"])
    # mp.spawn(train_worker, args=(world_size, config), nprocs=world_size, join=True)
    train_and_evaluate(config)
    # ensemble_train_and_evaluate(config)



"""
Author: Xinzhuo Wu
Date: 2023-12-29 14:20:18
LastEditTime: 2024-05-31 11:00:01
LastEditors: Wenyu Ouyang
Description: A simple evaluate model test
FilePath: \torchhydro\tests\test_evaluate_grid_lstm.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
import warnings
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.deep_hydro import DeepHydro
from torchhydro.trainers.trainer import set_random_seed
from torchhydro.trainers.resulter import Resulter

warnings.filterwarnings("ignore")


@pytest.fixture()
def config_data():
    project_name = "test_evalute_spp_lstm/ex1"
    train_path = os.path.join(os.getcwd(), "results", "test_spp_lstm", "ex1_0")
    args = cmd(
        sub=project_name,
        source="HydroGrid",
        source_path=[
            {
                "gpm": "basins-origin/hour_data/1h/grid_data/grid_gpm_data",
                "gfs": "basins-origin/hour_data/1h/grid_data/grid_gfs_data",
                "smap": "basins-origin/hour_data/1h/grid_data/grid_smap_data",
                "target": "basins-origin/hour_data/1h/mean_data/mean_data_target",
                "attributes": "basins-origin/attributes.nc",
            }
        ],
        ctx=[2],
        model_name="SPPLSTM2",
        model_hyperparam={
            "forecast_history": 168,
            "forecast_length": 24,
            "p_n_output": 1,
            "p_n_hidden_states": 60,
            "p_dropout": 0.25,
            "p_in_channels": 1,
            "p_out_channels": 8,
            "len_c": 15,
            "s_forecast_history": None,
            "s_n_output": 1,
            "s_n_hidden_states": 60,
            "s_dropout": 0.25,
            "s_in_channels": 8,
            "s_out_channels": 8,
        },
        gage_id=["21401550"],
        batch_size=256,
        var_t=[
            ["gpm_tp"],
            [
                "None",
            ],
            [
                "None",
            ],
        ],
        var_out=["streamflow"],
        var_c=[
            "area",  # 面积
            "ele_mt_smn",  # 海拔(空间平均)
            "slp_dg_sav",  # 地形坡度 (空间平均)
            "sgr_dk_sav",  # 河流坡度 (平均)
            "for_pc_sse",  # 森林覆盖率
            "glc_cl_smj",  # 土地覆盖类型
            "run_mm_syr",  # 陆面径流 (流域径流的空间平均值)
            "inu_pc_slt",  # 淹没范围 (长期最大)
            "cmi_ix_syr",  # 气候湿度指数
            "aet_mm_syr",  # 实际蒸散发 (年平均)
            "snw_pc_syr",  # 雪盖范围 (年平均)
            "swc_pc_syr",  # 土壤水含量
            "gwt_cm_sav",  # 地下水位深度
            "cly_pc_sav",  # 土壤中的黏土、粉砂、砂粒含量
            "dor_pc_pva",  # 调节程度
        ],
        dataset="GridDataset",
        sampler="HydroSampler",
        scaler="MutiBasinScaler",
        test_period=[
            ("2017-07-01", "2017-09-29"),
        ],
        rolling=False,
        weight_path=os.path.join(train_path, "best_model.pth"),
        stat_dict_file=os.path.join(train_path, "MutiBasinScaler_stat.json"),
        continue_train=False,
    )
    config_data = default_config_file()
    update_cfg(config_data, args)
    return config_data


def test_evaluate_spp_lstm(config_data):
    random_seed = config_data["training_cfgs"]["random_seed"]
    set_random_seed(random_seed)
    resulter = Resulter(config_data)
    model = DeepHydro(config_data)
    test_acc = model.model_evaluate()
    print("summary test_accuracy", test_acc[0])
    resulter.save_result(
        test_acc[0],
        test_acc[1],
    )



"""
Author: Wenyu Ouyang
Date: 2023-07-25 16:47:19
LastEditTime: 2024-05-27 16:22:15
LastEditors: Wenyu Ouyang
Description: Test a full training and evaluating process
FilePath: \torchhydro\tests\test_caravan_train.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
from torchhydro import SETTING
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate


@pytest.fixture()
def var_c():
    return ["area", "p_mean", "pet_mean", "aridity", "frac_snow", "moisture_index"]


@pytest.fixture()
def var_t():
    return [
        "surface_net_solar_radiation_mean",
        "total_precipitation_sum",
        "potential_evaporation_sum",
        "temperature_2m_max",
        "temperature_2m_min",
        "surface_pressure_mean",
    ]


@pytest.fixture()
def config(var_c, var_t):
    project_name = "test_caravan/exp5"
    config_data = default_config_file()
    args = cmd(
        sub=project_name,
        source="Caravan",
        source_path=os.path.join(
            SETTING["local_data_path"]["datasets-origin"], "caravan"
        ),
        source_region="Global",
        ctx=[-1],
        model_name="CpuLSTM",
        model_hyperparam={
            "n_input_features": len(var_t) + len(var_c),
            "n_output_features": 1,
            "n_hidden_states": 256,
        },
        gage_id=[
            "camels_01022500",
            "camels_01031500",
            "camels_01047000",
            "camels_01052500",
            "camels_01054200",
            "camels_01055000",
            "camels_01057000",
            "camelsaus_102101A",
            "camelsaus_108003A",
            "hysets_06444000",
        ],
        batch_size=8,
        forecast_history=0,
        forecast_length=20,
        var_t=var_t,
        var_c=var_c,
        var_out=["streamflow"],
        dataset="StreamflowDataset",
        sampler="KuaiSampler",
        scaler="DapengScaler",
        train_epoch=2,
        save_epoch=1,
        te=2,
        train_period=["2000-10-01", "2001-10-01"],
        valid_period=["2001-10-01", "2002-10-01"],
        test_period=["2002-10-01", "2003-10-01"],
        loss_func="RMSESum",
        opt="Adam",
        # key is epoch, start from 1
        lr_scheduler={1: 1e-2, 2: 5e-3, 3: 1e-3},
        which_first_tensor="sequence",
    )
    update_cfg(config_data, args)
    return config_data


def test_train_evaluate(config):
    train_and_evaluate(config)



"""
Author: Wenyu Ouyang
Date: 2024-04-09 21:16:45
LastEditTime: 2024-04-09 21:23:48
LastEditors: Wenyu Ouyang
Description: Test the data utilities
FilePath: \torchhydro\tests\test_data_utils.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import numpy as np
import pytest
import xarray as xr

from torchhydro.datasets.data_utils import warn_if_nan


def test_warn_if_nan_no_nan_values():
    # Create a dataarray without any NaN values
    data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    da = xr.DataArray(data)
    # Call the function and assert that it returns False
    assert not warn_if_nan(da)


def test_warn_if_nan_with_nan_values():
    # Create a dataarray with NaN values
    data = np.array([[1, np.nan, 3], [4, 5, np.nan], [7, 8, 9]])
    da = xr.DataArray(data)
    # Call the function and assert that it returns True
    assert warn_if_nan(da)


def test_warn_if_nan_all_nan_values():
    # Create a dataarray with all NaN values
    data = np.array(
        [[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]
    )
    da = xr.DataArray(data)
    # Call the function and assert that it raises a ValueError
    with pytest.raises(ValueError):
        warn_if_nan(da, nan_mode="all")


def test_warn_if_nan_max_display():
    # Create a dataarray with NaN values
    data = np.array([[1, np.nan, 3], [4, 5, np.nan], [7, np.nan, 9]])
    da = xr.DataArray(data)
    # Call the function with a max_display of 2
    # Assert that it raises a warning with the correct message
    with pytest.warns(
        UserWarning,
        match=r"The dataarray contains 3 NaN values! Here are the indices of the first 2 NaNs:",
    ):
        warn_if_nan(da, max_display=2)



"""
Author: Wenyu Ouyang
Date: 2021-12-31 11:08:29
LastEditTime: 2024-05-04 11:30:00
LastEditors: Wenyu Ouyang
Description: Training function for DL models
FilePath: \torchhydro\torchhydro\trainers\train_logger.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

from contextlib import contextmanager
from datetime import datetime
import json
import os
import time
from hydroutils import hydro_file
import numpy as np
import torch
from torch.utils.tensorboard import SummaryWriter


def save_model(model, model_file, gpu_num=1):
    try:
        if torch.cuda.device_count() > 1 and gpu_num > 1:
            torch.save(model.module.state_dict(), model_file)
        else:
            torch.save(model.state_dict(), model_file)
    except RuntimeError:
        torch.save(model.module.state_dict(), model_file)


def save_model_params_log(params, params_log_path):
    time_stamp = datetime.now().strftime("%d_%B_%Y%I_%M%p")
    params_log_file = os.path.join(params_log_path, f"{time_stamp}.json")
    hydro_file.serialize_json(params, params_log_file)


class TrainLogger:
    def __init__(self, model_filepath, params, opt):
        self.training_cfgs = params["training_cfgs"]
        self.data_cfgs = params["data_cfgs"]
        self.evaluation_cfgs = params["evaluation_cfgs"]
        self.model_cfgs = params["model_cfgs"]
        self.opt = opt
        self.training_save_dir = model_filepath
        self.tb = SummaryWriter(self.training_save_dir)
        self.session_params = []
        self.train_time = []
        # log loss for each epoch
        self.epoch_loss = []

    def save_session_param(
        self, epoch, total_loss, n_iter_ep, valid_loss=None, valid_metrics=None
    ):
        if valid_loss is None or valid_metrics is None:
            epoch_params = {
                "epoch": epoch,
                "train_loss": str(total_loss),
                "iter_num": n_iter_ep,
            }
        else:
            epoch_params = {
                "epoch": epoch,
                "train_loss": str(total_loss),
                "validation_loss": str(valid_loss),
                "validation_metric": valid_metrics,
                "iter_num": n_iter_ep,
            }
        epoch_params["train_time"] = self.train_time[epoch - 1]
        self.session_params.append(epoch_params)

    @contextmanager
    def log_epoch_train(self, epoch):
        start_time = time.time()
        logs = {}
        # here content in the with block will be performed
        yield logs
        total_loss = logs["train_loss"]
        elapsed_time = time.time() - start_time
        lr = self.opt.param_groups[0]["lr"]
        log_str = "Epoch {} Loss {:.4f} time {:.2f} lr {}".format(
            epoch, total_loss, elapsed_time, lr
        )
        print(log_str)
        model = logs["model"]
        print(model)
        self.tb.add_scalar("Loss", total_loss, epoch)
        # self.plot_hist_img(model, epoch)
        self.train_time.append(log_str)
        self.epoch_loss.append(total_loss)

    @contextmanager
    def log_epoch_valid(self, epoch):
        logs = {}
        yield logs
        valid_loss = logs["valid_loss"]
        if self.evaluation_cfgs["calc_metrics"]:
            valid_metrics = logs["valid_metrics"]
            val_log = "Epoch {} Valid Loss {:.4f} Valid Metric {}".format(
                epoch, valid_loss, valid_metrics
            )
            print(val_log)
            self.tb.add_scalar("ValidLoss", valid_loss, epoch)
            target_col = self.data_cfgs["target_cols"]
            evaluation_metrics = self.evaluation_cfgs["metrics"]
            for i in range(len(target_col)):
                for evaluation_metric in evaluation_metrics:
                    self.tb.add_scalar(
                        f"Valid{target_col[i]}{evaluation_metric}mean",
                        np.mean(
                            valid_metrics[f"{evaluation_metric} of {target_col[i]}"]
                        ),
                        epoch,
                    )
                    self.tb.add_scalar(
                        f"Valid{target_col[i]}{evaluation_metric}median",
                        np.median(
                            valid_metrics[f"{evaluation_metric} of {target_col[i]}"]
                        ),
                        epoch,
                    )
        else:
            val_log = "Epoch {} Valid Loss {:.4f} ".format(epoch, valid_loss)
            print(val_log)
            self.tb.add_scalar("ValidLoss", valid_loss, epoch)

    def save_model_and_params(self, model, epoch, params):
        final_epoch = params["training_cfgs"]["epochs"]
        save_epoch = params["training_cfgs"]["save_epoch"]
        if save_epoch is None or save_epoch == 0 and epoch != final_epoch:
            return
        if (save_epoch > 0 and epoch % save_epoch == 0) or epoch == final_epoch:
            # save for save_epoch
            model_file = os.path.join(
                self.training_save_dir, f"model_Ep{str(epoch)}.pth"
            )
            save_model(model, model_file)
        if epoch == final_epoch:
            self._save_final_epoch(params, model)

    def _save_final_epoch(self, params, model):
        # In final epoch, we save the model and params in test_path
        final_path = params["data_cfgs"]["test_path"]
        params["run"] = self.session_params
        time_stamp = datetime.now().strftime("%d_%B_%Y%I_%M%p")
        model_save_path = os.path.join(final_path, f"{time_stamp}_model.pth")
        save_model(model, model_save_path)
        save_model_params_log(params, final_path)
        # also save one for a training directory for one hyperparameter setting
        save_model_params_log(params, self.training_save_dir)

    def plot_hist_img(self, model, global_step):
        for tag, parm in model.named_parameters():
            self.tb.add_histogram(
                f"{tag}_hist", parm.detach().cpu().numpy(), global_step
            )
            if len(parm.shape) == 2:
                img_format = "HW"
                if parm.shape[0] > parm.shape[1]:
                    img_format = "WH"
                    self.tb.add_image(
                        f"{tag}_img",
                        parm.detach().cpu().numpy(),
                        global_step,
                        dataformats=img_format,
                    )

    def plot_model_structure(self, model):
        """plot model structure in tensorboard

        Parameters
        ----------
        model :
            torch model
        """
        # input4modelplot = torch.randn(
        #     self.data_cfgs["batch_size"],
        #     self.data_cfgs["forecast_history"],
        #     # self.model_cfgs["model_hyperparam"]["n_input_features"],
        #     self.model_cfgs["model_hyperparam"]["input_size"],
        # )
        if self.data_cfgs["model_mode"] == "single":
            input4modelplot = [
                torch.randn(
                    self.data_cfgs["batch_size"],
                    self.data_cfgs["forecast_history"],
                    self.data_cfgs["input_features"] - 1,
                ),
                torch.randn(
                    self.data_cfgs["batch_size"],
                    self.data_cfgs["forecast_history"],
                    self.data_cfgs["cnn_size"],
                ),
                torch.rand(
                    self.data_cfgs["batch_size"], 1, self.data_cfgs["output_features"]
                ),
            ]
        else:
            input4modelplot = [
                torch.randn(
                    self.data_cfgs["batch_size"],
                    self.data_cfgs["forecast_history"],
                    self.data_cfgs["input_features"],
                ),
                torch.randn(
                    self.data_cfgs["batch_size"],
                    self.data_cfgs["forecast_history"],
                    self.data_cfgs["input_size_encoder2"],
                ),
                torch.rand(
                    self.data_cfgs["batch_size"], 1, self.data_cfgs["output_features"]
                ),
            ]
        self.tb.add_graph(model, input4modelplot)



"""
Author: Wenyu Ouyang
Date: 2023-09-25 08:21:27
LastEditTime: 2024-05-27 15:59:09
LastEditors: Wenyu Ouyang
Description: Some sampling class or functions
FilePath: \torchhydro\torchhydro\datasets\sampler.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

from collections import defaultdict
import numpy as np
from torch.utils.data import RandomSampler, Sampler
from torchhydro.datasets.data_sets import BaseDataset
from typing import Iterator, Optional, Sized
import torch


class KuaiSampler(RandomSampler):
    def __init__(
        self,
        data_source,
        batch_size,
        warmup_length,
        rho_horizon,
        ngrid,
        nt,
    ):
        """a sampler from Kuai Fang's paper: https://doi.org/10.1002/2017GL075619
           He used a random pick-up that we don't need to iterate all samples.
           Then, we can train model more quickly

        Parameters
        ----------
        data_source : torch.utils.data.Dataset
            just a object of dataset class inherited from torch.utils.data.Dataset
        batch_size : int
            we need batch_size to calculate the number of samples in an epoch
        warmup_length : int
            warmup length, typically for physical hydrological models
        rho_horizon : int
            sequence length of a mini-batch, for encoder-decoder models, rho+horizon, for decoder-only models, horizon
        ngrid : int
            number of basins
        nt : int
            number of all periods
        """
        while batch_size * rho_horizon >= ngrid * nt:
            # try to use a smaller batch_size to make the model runnable
            batch_size = int(batch_size / 10)
        batch_size = max(batch_size, 1)
        # 99% chance that all periods' data are used in an epoch
        n_iter_ep = int(
            np.ceil(
                np.log(0.01)
                / np.log(1 - batch_size * rho_horizon / ngrid / (nt - warmup_length))
            )
        )
        assert n_iter_ep >= 1
        # __len__ means the number of all samples, then, the number of loops in an epoch is __len__()/batch_size = n_iter_ep
        # hence we return n_iter_ep * batch_size
        num_samples = n_iter_ep * batch_size
        super(KuaiSampler, self).__init__(data_source, num_samples=num_samples)


class HydroSampler(Sampler[int]):
    """
    A custom sampler for hydrological modeling that iterates over a dataset in
    a way tailored for batches of hydrological data. It ensures that each batch
    contains data from a single randomly selected 'basin' out of several basins,
    with batches constructed to respect the specified batch size and the unique
    characteristics of hydrological datasets.

    Parameters:
    - data_source (Sized): The dataset to sample from, expected to have a `data_cfgs` attribute.
    - num_samples (Optional[int], default=None): The total number of samples to draw (optional).
    - generator: A PyTorch Generator object for random number generation (optional).

    The sampler divides the dataset by the number of basins, then iterates through
    each basin's range in shuffled order, ensuring non-overlapping, basin-specific
    batches suitable for models that predict hydrological outcomes.
    """

    data_source: Sized

    def __init__(
        self,
        data_source: Sized,
        num_samples: Optional[int] = None,
        generator=None,
    ) -> None:
        self.data_source = data_source
        self._num_samples = num_samples
        self.generator = generator

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError(
                f"num_samples should be a positive integer value, but got num_samples={self.num_samples}"
            )

    @property
    def num_samples(self) -> int:
        return len(self.data_source)

    def __iter__(self) -> Iterator[int]:
        n = self.data_source.data_cfgs["batch_size"]
        basin_number = len(self.data_source.data_cfgs["object_ids"])
        basin_range = len(self.data_source) // basin_number
        if n > basin_range:
            raise ValueError(
                f"batch_size should equal or less than basin_range={basin_range} "
            )

        if self.generator is None:
            seed = int(torch.empty((), dtype=torch.int64).random_().item())
            generator = torch.Generator()
            generator.manual_seed(seed)
        else:
            generator = self.generator

        # basin_list = torch.randperm(basin_number)
        # for select_basin in basin_list:
        #     x = torch.randperm(basin_range)
        #     for i in range(0, basin_range, n):
        #         yield from (x[i : i + n] + basin_range * select_basin.item()).tolist()
        x = torch.randperm(self.num_samples)
        for i in range(0, self.num_samples, n):
            yield from (x[i : i + n]).tolist()

    def __len__(self) -> int:
        return self.num_samples


def fl_sample_basin(dataset: BaseDataset):
    """
    Sample one basin data as a client from a dataset for federated learning

    Parameters
    ----------
    dataset
        dataset

    Returns
    -------
        dict of image index
    """
    lookup_table = dataset.lookup_table
    basins = dataset.basins
    # one basin is one user
    num_users = len(basins)
    # set group for basins
    basin_groups = defaultdict(list)
    for idx, (basin, date) in lookup_table.items():
        basin_groups[basin].append(idx)

    # one user is one basin
    user_basins = defaultdict(list)
    for i, basin in enumerate(basins):
        user_id = i % num_users
        user_basins[user_id].append(basin)

    # a lookup_table subset for each user
    user_lookup_tables = {}
    for user_id, basins in user_basins.items():
        user_lookup_table = {}
        for basin in basins:
            for idx in basin_groups[basin]:
                user_lookup_table[idx] = lookup_table[idx]
        user_lookup_tables[user_id] = user_lookup_table

    return user_lookup_tables


def fl_sample_region(dataset: BaseDataset):
    """
    Sample one region data as a client from a dataset for federated learning

    TODO: not finished

    """
    num_users = 10
    num_shards, num_imgs = 200, 250
    idx_shard = list(range(num_shards))
    dict_users = {i: np.array([]) for i in range(num_users)}
    idxs = np.arange(num_shards * num_imgs)
    # labels = dataset.train_labels.numpy()
    labels = np.array(dataset.train_labels)

    # sort labels
    idxs_labels = np.vstack((idxs, labels))
    idxs_labels = idxs_labels[:, idxs_labels[1, :].argsort()]
    idxs = idxs_labels[0, :]

    # divide and assign
    for i in range(num_users):
        rand_set = set(np.random.choice(idx_shard, 2, replace=False))
        idx_shard = list(set(idx_shard) - rand_set)
        for rand in rand_set:
            dict_users[i] = np.concatenate(
                (dict_users[i], idxs[rand * num_imgs : (rand + 1) * num_imgs]), axis=0
            )
    return dict_users



import os
import time
import numpy as np
import pandas as pd
import torch
import xarray as xr
import fnmatch
from typing import Tuple
from functools import reduce
from torch.utils.data import DataLoader
from torch.nn import functional as F
from tbparse import SummaryReader

from hydroutils.hydro_stat import stat_error
from hydrodatasource.utils.utils import streamflow_unit_conv

from torchhydro.configs.model_config import MODEL_PARAM_TEST_WAY
from torchhydro.datasets.data_sources import data_sources_dict
from torchhydro.trainers.train_logger import save_model_params_log
from torchhydro.explainers.shap import (
    deep_explain_model_heatmap,
    deep_explain_model_summary_plot,
    shap_summary_plot,
)
from torchhydro.trainers.train_utils import (
    calculate_and_record_metrics,
    cellstates_when_inference,
    read_pth_from_model_loader,
    model_infer,
)
from torchhydro.trainers.deep_hydro import DeepHydro


def set_unit_to_var(ds):
    units_dict = ds.attrs["units"]
    for var_name, units in units_dict.items():
        if var_name in ds:
            ds[var_name].attrs["units"] = units
    if "units" in ds.attrs:
        del ds.attrs["units"]
    return ds


class Resulter:
    def __init__(self, cfgs) -> None:
        self.cfgs = cfgs
        self.result_dir = cfgs["data_cfgs"]["test_path"]
        if not os.path.exists(self.result_dir):
            os.makedirs(self.result_dir)

    @property
    def pred_name(self):
        return f"epoch{str(self.chosen_trained_epoch)}flow_pred"

    @property
    def obs_name(self):
        return f"epoch{str(self.chosen_trained_epoch)}flow_obs"

    @property
    def chosen_trained_epoch(self):
        model_loader = self.cfgs["evaluation_cfgs"]["model_loader"]
        if model_loader["load_way"] == "specified":
            epoch_name = str(model_loader["test_epoch"])
        elif model_loader["load_way"] == "best":
            # NOTE: TO make it consistent with the name in case of model_loader["load_way"] == "pth", the name have to be "best_model.pth"
            epoch_name = "best_model.pth"
        elif model_loader["load_way"] == "latest":
            epoch_name = str(self.cfgs["training_cfgs"]["epochs"])
        elif model_loader["load_way"] == "pth":
            epoch_name = model_loader["pth_path"].split(os.sep)[-1]
        else:
            raise ValueError("Invalid load_way")
        return epoch_name

    def save_cfg(self, cfgs):
        # save the cfgs after training
        # update the cfgs with the latest one
        self.cfgs = cfgs
        param_file_exist = any(
            (
                fnmatch.fnmatch(file, "*.json")
                and "_stat" not in file  # statistics json file
                and "_dict" not in file  # data cache json file
            )
            for file in os.listdir(self.result_dir)
        )
        if not param_file_exist:
            # although we save params log during training, but sometimes we directly evaluate a model
            # so here we still save params log if param file does not exist
            # no param file was saved yet, here we save data and params setting
            save_model_params_log(cfgs, self.result_dir)

    def save_result(self, pred, obs):
        """
        save the pred value of testing period and obs value

        Parameters
        ----------
        pred
            predictions
        obs
            observations
        pred_name
            the file name of predictions
        obs_name
            the file name of observations

        Returns
        -------
        None
        """
        save_dir = self.result_dir
        flow_pred_file = os.path.join(save_dir, self.pred_name)
        flow_obs_file = os.path.join(save_dir, self.obs_name)
        pred = set_unit_to_var(pred)
        obs = set_unit_to_var(obs)
        pred.to_netcdf(flow_pred_file + ".nc")
        obs.to_netcdf(flow_obs_file + ".nc")

    def eval_result(self, preds_xr, obss_xr):
        # types of observations
        target_col = self.cfgs["data_cfgs"]["target_cols"]
        evaluation_metrics = self.cfgs["evaluation_cfgs"]["metrics"]
        basin_ids = self.cfgs["data_cfgs"]["object_ids"]
        test_path = self.cfgs["data_cfgs"]["test_path"]
        # Assume object_ids like ['changdian_61561']
        # fill_nan: "no" means ignoring the NaN value;
        #           "sum" means calculate the sum of the following values in the NaN locations.
        #           For example, observations are [1, nan, nan, 2], and predictions are [0.3, 0.3, 0.3, 1.5].
        #           Then, "no" means [1, 2] v.s. [0.3, 1.5] while "sum" means [1, 2] v.s. [0.3 + 0.3 + 0.3, 1.5].
        #           If it is a str, then all target vars use same fill_nan method;
        #           elif it is a list, each for a var
        fill_nan = self.cfgs["evaluation_cfgs"]["fill_nan"]
        #  Then evaluate the model metrics
        if type(fill_nan) is list and len(fill_nan) != len(target_col):
            raise ValueError("length of fill_nan must be equal to target_col's")
        for i, col in enumerate(target_col):
            eval_log = {}
            obs = obss_xr[col].to_numpy()
            pred = preds_xr[col].to_numpy()

            eval_log = calculate_and_record_metrics(
                obs,
                pred,
                evaluation_metrics,
                col,
                fill_nan[i] if isinstance(fill_nan, list) else fill_nan,
                eval_log,
            )
            # Create pandas DataFrames from eval_log for each target variable (e.g., streamflow)
            # Create a dictionary to hold the data for the DataFrame
            data = {}
            # Iterate over metrics in eval_log
            for metric, values in eval_log.items():
                # Remove 'of streamflow' (or similar) from the metric name
                clean_metric = metric.replace(f"of {col}", "").strip()

                # Add the cleaned metric to the data dictionary
                data[clean_metric] = values

            # Create a DataFrame using object_ids as the index and metrics as columns
            df = pd.DataFrame(data, index=basin_ids)

            # Define the output file name based on the target variable
            output_file = os.path.join(test_path, f"metric_{col}.csv")

            # Save the DataFrame to a CSV file
            df.to_csv(output_file, index_label="basin_id")

        # Finally, try to explain model behaviour using shap
        is_shap = self.cfgs["evaluation_cfgs"]["explainer"] == "shap"
        if is_shap:
            shap_summary_plot(self.model, self.traindataset, self.testdataset)
            # deep_explain_model_summary_plot(self.model, test_data)
            # deep_explain_model_heatmap(self.model, test_data)

    def _convert_streamflow_units(self, ds):
        """convert the streamflow units to m^3/s

        Parameters
        ----------
        pred : np.array
            predictions

        Returns
        -------
        """
        data_cfgs = self.cfgs["data_cfgs"]
        source_name = data_cfgs["source_cfgs"]["source_name"]
        source_path = data_cfgs["source_cfgs"]["source_path"]
        other_settings = data_cfgs["source_cfgs"].get("other_settings", {})
        data_source = data_sources_dict[source_name](source_path, **other_settings)
        basin_id = data_cfgs["object_ids"]
        # NOTE: all datasource should have read_area method
        basin_area = data_source.read_area(basin_id)
        target_unit = "m^3/s"
        # NOTE: the name of var flow should be streamflow
        var_flow = "streamflow"
        streamflow_ds = ds[[var_flow]]
        ds_ = streamflow_unit_conv(
            streamflow_ds, basin_area, target_unit=target_unit, inverse=True
        )
        new_ds = ds.copy(deep=True)
        new_ds[var_flow] = ds_[var_flow]
        return new_ds

    def load_result(self, convert_flow_unit=False) -> Tuple[np.array, np.array]:
        """load the pred value of testing period and obs value"""
        save_dir = self.result_dir
        pred_file = os.path.join(save_dir, self.pred_name + ".nc")
        obs_file = os.path.join(save_dir, self.obs_name + ".nc")
        pred = xr.open_dataset(pred_file)
        obs = xr.open_dataset(obs_file)
        if convert_flow_unit:
            pred = self._convert_streamflow_units(pred)
            obs = self._convert_streamflow_units(obs)
        return pred, obs

    def save_intermediate_results(self, **kwargs):
        """Load model weights and deal with some intermediate results"""
        is_cell_states = kwargs.get("is_cell_states", False)
        is_pbm_params = kwargs.get("is_pbm_params", False)
        cfgs = self.cfgs
        cfgs["training_cfgs"]["train_mode"] = False
        training_cfgs = cfgs["training_cfgs"]
        seq_first = training_cfgs["which_first_tensor"] == "sequence"
        if is_cell_states:
            # TODO: not support return_cell_states yet
            return cellstates_when_inference(seq_first, data_cfgs, pred)
        if is_pbm_params:
            self._save_pbm_params(cfgs, seq_first)

    def _save_pbm_params(self, cfgs, seq_first):
        training_cfgs = cfgs["training_cfgs"]
        model_loader = cfgs["evaluation_cfgs"]["model_loader"]
        model_pth_dir = cfgs["data_cfgs"]["test_path"]
        weight_path = read_pth_from_model_loader(model_loader, model_pth_dir)
        cfgs["model_cfgs"]["weight_path"] = weight_path
        cfgs["training_cfgs"]["device"] = [0] if torch.cuda.is_available() else [-1]
        deephydro = DeepHydro(cfgs)
        device = deephydro.device
        dl_model = deephydro.model.dl_model
        pb_model = deephydro.model.pb_model
        param_func = deephydro.model.param_func
        # TODO: check for dplnnmodule model
        param_test_way = deephydro.model.param_test_way
        test_dataloader = DataLoader(
            deephydro.testdataset,
            batch_size=training_cfgs["batch_size"],
            shuffle=False,
            sampler=None,
            batch_sampler=None,
            drop_last=False,
            timeout=0,
            worker_init_fn=None,
        )
        deephydro.model.eval()
        # here the batch is just an index of lookup table, so any batch size could be chosen
        params_lst = []
        with torch.no_grad():
            for xs, ys in test_dataloader:
                ys, gen = model_infer(seq_first, device, dl_model, xs[1], ys)
                # we set all params' values in [0, 1] and will scale them when forwarding
                if param_func == "clamp":
                    params_ = torch.clamp(gen, min=0.0, max=1.0)
                elif param_func == "sigmoid":
                    params_ = F.sigmoid(gen)
                else:
                    raise NotImplementedError(
                        "We don't provide this way to limit parameters' range!! Please choose sigmoid or clamp"
                    )
                # just get one-period values, here we use the final period's values
                params = params_[:, -1, :]
                params_lst.append(params)
        pb_params = reduce(lambda a, b: torch.cat((a, b), dim=0), params_lst)
        # trans tensor to pandas dataframe
        sites = deephydro.cfgs["data_cfgs"]["object_ids"]
        params_names = pb_model.params_names
        params_df = pd.DataFrame(
            pb_params.cpu().numpy(), columns=params_names, index=sites
        )
        save_param_file = os.path.join(
            model_pth_dir, f"pb_params_{int(time.time())}.csv"
        )
        params_df.to_csv(save_param_file, index_label="GAGE_ID")

    def read_tensorboard_log(self, **kwargs):
        """read tensorboard log files"""
        is_scalar = kwargs.get("is_scalar", False)
        is_histogram = kwargs.get("is_histogram", False)
        log_dir = self.cfgs["data_cfgs"]["test_path"]
        if is_scalar:
            scalar_file = os.path.join(log_dir, "tb_scalars.csv")
            if not os.path.exists(scalar_file):
                reader = SummaryReader(log_dir)
                df_scalar = reader.scalars
                df_scalar.to_csv(scalar_file, index=False)
            else:
                df_scalar = pd.read_csv(scalar_file)
        if is_histogram:
            histogram_file = os.path.join(log_dir, "tb_histograms.csv")
            if not os.path.exists(histogram_file):
                reader = SummaryReader(log_dir)
                df_histogram = reader.histograms
                df_histogram.to_csv(histogram_file, index=False)
            else:
                df_histogram = pd.read_csv(histogram_file)
        if is_scalar and is_histogram:
            return df_scalar, df_histogram
        elif is_scalar:
            return df_scalar
        elif is_histogram:
            return df_histogram

    # TODO: the following code is not finished yet
    def load_ensemble_result(
        self, save_dirs, test_epoch, flow_unit="m3/s", basin_areas=None
    ) -> Tuple[np.array, np.array]:
        """
        load ensemble mean value

        Parameters
        ----------
        save_dirs
        test_epoch
        flow_unit
            default is m3/s, if it is not m3/s, transform the results
        basin_areas
            if unit is mm/day it will be used, default is None

        Returns
        -------

        """
        preds = []
        obss = []
        for save_dir in save_dirs:
            pred_i, obs_i = self.load_result(save_dir, test_epoch)
            if pred_i.ndim == 3 and pred_i.shape[-1] == 1:
                pred_i = pred_i.reshape(pred_i.shape[0], pred_i.shape[1])
                obs_i = obs_i.reshape(obs_i.shape[0], obs_i.shape[1])
            preds.append(pred_i)
            obss.append(obs_i)
        preds_np = np.array(preds)
        obss_np = np.array(obss)
        pred_mean = np.mean(preds_np, axis=0)
        obs_mean = np.mean(obss_np, axis=0)
        if flow_unit == "mm/day":
            if basin_areas is None:
                raise ArithmeticError("No basin areas we cannot calculate")
            basin_areas = np.repeat(basin_areas, obs_mean.shape[1], axis=0).reshape(
                obs_mean.shape
            )
            obs_mean = obs_mean * basin_areas * 1e-3 * 1e6 / 86400
            pred_mean = pred_mean * basin_areas * 1e-3 * 1e6 / 86400
        elif flow_unit == "m3/s":
            pass
        elif flow_unit == "ft3/s":
            obs_mean = obs_mean / 35.314666721489
            pred_mean = pred_mean / 35.314666721489
        return pred_mean, obs_mean

    def eval_ensemble_result(
        self,
        save_dirs,
        test_epoch,
        return_value=False,
        flow_unit="m3/s",
        basin_areas=None,
    ) -> Tuple[np.array, np.array]:
        """calculate statistics for ensemble results

        Parameters
        ----------
        save_dirs : _type_
            where the results save
        test_epoch : _type_
            we name the results files with the test_epoch
        return_value : bool, optional
            if True, return (inds_df, pred_mean, obs_mean), by default False
        flow_unit : str, optional
            arg for load_ensemble_result, by default "m3/s"
        basin_areas : _type_, optional
            arg for load_ensemble_result, by default None

        Returns
        -------
        Tuple[np.array, np.array]
            inds_df or (inds_df, pred_mean, obs_mean)
        """
        pred_mean, obs_mean = self.load_ensemble_result(
            save_dirs, test_epoch, flow_unit=flow_unit, basin_areas=basin_areas
        )
        inds = stat_error(obs_mean, pred_mean)
        inds_df = pd.DataFrame(inds)
        return (inds_df, pred_mean, obs_mean) if return_value else inds_df



"""
Author: Wenyu Ouyang
Date: 2023-07-31 08:40:43
LastEditTime: 2024-09-10 20:10:25
LastEditors: Wenyu Ouyang
Description: Top-level package for torchhydro
FilePath: \torchhydro\torchhydro\__init__.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

""""""

__author__ = """Wenyu Ouyang"""
__email__ = "wenyuouyang@outlook.com"
__version__ = '0.0.8'

from pathlib import Path
import yaml
import os

from hydroutils import hydro_file

CACHE_DIR = Path(hydro_file.get_cache_dir())
SETTING_FILE = os.path.join(Path.home(), "hydro_setting.yml")


def read_setting(setting_path):
    if not os.path.exists(setting_path):
        raise FileNotFoundError(f"Configuration file not found: {setting_path}")

    with open(setting_path, "r", encoding="utf-8") as file:
        setting = yaml.safe_load(file)

    example_setting = (
        "minio:\n"
        "  server_url: 'http://minio.waterism.com:9090' # Update with your URL\n"
        "  client_endpoint: 'http://minio.waterism.com:9000' # Update with your URL\n"
        "  access_key: 'your minio access key'\n"
        "  secret: 'your minio secret'\n\n"
        "local_data_path:\n"
        "  root: 'D:\\data\\waterism' # Update with your root data directory\n"
        "  datasets-origin: 'D:\\data\\waterism\\datasets-origin'\n"
        "  datasets-interim: 'D:\\data\\waterism\\datasets-interim'"
        "postgres:\n"
        "  server_url: 'your database server URL'"  # if you don't have a database, you can just write this key and leave the rest empty
        "  port: your database port"
        "  username: 'your database username'"
        "  password: 'your database password'"
        "  database: 'your database name'"
    )

    if setting is None:
        raise ValueError(
            f"Configuration file is empty or has invalid format.\n\nExample configuration:\n{example_setting}"
        )

    # Define the expected structure
    expected_structure = {
        "minio": ["server_url", "client_endpoint", "access_key", "secret"],
        "local_data_path": ["root", "datasets-origin", "datasets-interim"],
        "postgres": ["server_url", "port", "username", "password", "database"],
    }

    # Validate the structure
    try:
        for key, subkeys in expected_structure.items():
            if key not in setting:
                raise KeyError(f"Missing required key in config: {key}")

            if isinstance(subkeys, list):
                for subkey in subkeys:
                    if subkey not in setting[key]:
                        raise KeyError(f"Missing required subkey '{subkey}' in '{key}'")
    except KeyError as e:
        raise ValueError(
            f"Incorrect configuration format: {e}\n\nExample configuration:\n{example_setting}"
        ) from e

    return setting


try:
    SETTING = read_setting(SETTING_FILE)
except ValueError as e:
    print(e)
except Exception as e:
    print(f"Unexpected error: {e}")


from .configs import *
from .datasets import *
from .models import *
from .trainers import *
from .explainers import *



"""
Author: silencesoup silencesoup@outlook.com
Date: 2024-04-20 11:30:16
LastEditors: silencesoup silencesoup@outlook.com
LastEditTime: 2024-04-20 11:41:05
FilePath: /torchhydro/tests/test_model.py
"""

import torch
from torchhydro.models import seq2seq
from torchhydro.models import model_dict_function


def test_model():
    # put your model param here, it's just an example for seq2seq model
    model_configs = {
        "Seq2Seq": {
            "input_size": 16,
            "output_size": 1,
            "hidden_size": 256,
            "forecast_length": 24,
            "cnn_size": 120,
            "model_mode": "single",
        },
        "Seq2Seq_dual": {
            "input_size": 16,
            "output_size": 1,
            "hidden_size": 256,
            "forecast_length": 24,
            "cnn_size": 120,
            "model_mode": "dual",
            "input_size_encoder2": 1,
        },
    }
    batch_size = 2
    forecast_history = 168

    # Initialize the model
    model_config = model_configs["Seq2Seq"]

    model = model_dict_function.pytorch_model_dict["Seq2Seq"](
        input_size=model_config["input_size"],
        output_size=model_config["output_size"],
        hidden_size=model_config["hidden_size"],
        forecast_length=model_config["forecast_length"],
        cnn_size=model_config["cnn_size"],
        model_mode=model_config["model_mode"],
    )

    # Generate random inputs for testing
    # sourcery skip: no-conditionals-in-tests
    if model_config["model_mode"] == "single":
        src1 = torch.rand(
            batch_size,
            forecast_history,
            model_config["input_size"] - 1,
        )
        src2 = torch.rand(
            batch_size, forecast_history, model_config["cnn_size"]
        )
    else:  # "dual"
        src1 = torch.rand(
            batch_size, forecast_history, model_config["input_size"]
        )
        src2 = torch.rand(
            batch_size,
            forecast_history,
            model_config["input_size_encoder2"],
        )

    trg_start_token = torch.rand(batch_size, 1, model_config["output_size"])

    # Execute the model
    outputs = model(src1, src2, trg_start_token)  # Adjusted to pass parameters directly

    # Print outputs to verify
    print("Outputs shape:", outputs.shape)
    assert outputs.shape == (
        batch_size,
        model_config["forecast_length"],
        model_config["output_size"],
    ), "Output shape is incorrect"

    print("Test passed with output shape:", outputs.shape)



"""
Author: Wenyu Ouyang
Date: 2023-10-05 16:16:48
LastEditTime: 2024-02-14 16:12:50
LastEditors: Wenyu Ouyang
Description: A test function for transfer learning
FilePath: \torchhydro\tests\test_transfer_learning.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
import logging
import pandas as pd
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate, ensemble_train_and_evaluate

logging.basicConfig(level=logging.INFO)
for logger_name in logging.root.manager.loggerDict:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

show = pd.read_csv("data/basin_id(46+1).csv", dtype={"id": str})
gage_id = show["id"].values.tolist()


@pytest.fixture()
def config():
    weight_path = os.path.join(
        os.getcwd(), "results", "test_mean_seq2seq", "ex22", "best_model.pth"
    )
    project_name = "test_pretrain_enhanced/ex1"
    args = cmd(
        sub=project_name,
        source_cfgs={
            "source": "HydroMean",
            "source_path": {
                "forcing": "basins-origin/hour_data/1h/mean_data/data_forcing_gpm",
                "target": "basins-origin/hour_data/1h/mean_data/streamflow_basin",
                "attributes": "basins-origin/attributes.nc",
            },
        },
        ctx=[1],
        model_type="TransLearn",
        model_name="DataEnhanced",
        model_hyperparam={
            "hidden_length": 256,
            "input_size": 19,
            "output_size": 1,
            "hidden_size": 256,
            "cnn_size": 120,
            "forecast_length": 24,
            "model_mode": "dual",
            "prec_window": 1,
        },
        model_loader={"load_way": "best"},
        opt="Adam",
        lr_scheduler={
            "lr": 0.001,
            "lr_factor": 0.96,
        },
        loss_func="RMSESum",
        batch_size=256,
        forecast_history=336,
        forecast_length=24,
        train_period=[
            ("2015-06-01", "2015-09-30"),
            ("2016-06-01", "2016-09-30"),
            ("2017-06-01", "2017-09-30"),
            ("2018-06-01", "2018-09-30"),
            ("2019-06-01", "2019-09-30"),
            ("2020-06-01", "2020-09-30"),
            ("2021-06-01", "2021-09-30"),
            ("2022-06-01", "2022-09-30"),
        ],
        test_period=[
            ("2023-06-01", "2023-09-30"),
        ],
        valid_period=[
            ("2023-06-01", "2023-09-30"),  # 目前只支持一个时段
        ],
        dataset="MultiSourceDataset",
        sampler="HydroSampler",
        scaler="DapengScaler",
        weight_path=weight_path,
        weight_path_add={
            "freeze_params": [
                "encoder1.lstm.weight_ih_l0",
                "encoder1.lstm.weight_hh_l0",
                "encoder1.lstm.bias_ih_l0",
                "encoder1.lstm.bias_hh_l0",
                "encoder1.fc.weight",
                "encoder1.fc.bias",
                "encoder1.fc2.weight",
                "encoder1.fc2.bias",
                "encoder2.lstm.weight_ih_l0",
                "encoder2.lstm.weight_hh_l0",
                "encoder2.lstm.bias_ih_l0",
                "encoder2.lstm.bias_hh_l0",
                "encoder2.fc.weight",
                "encoder2.fc.bias",
                "encoder2.fc2.weight",
                "encoder2.fc2.bias",
                "decoder1.lstm.weight_ih_l0",
                "decoder1.lstm.weight_hh_l0",
                "decoder1.lstm.bias_ih_l0",
                "decoder1.lstm.bias_hh_l0",
                "decoder1.fc_out.weight",
                "decoder1.fc_out.bias",
                "decoder1.attention.attn.weight",
                "decoder2.lstm.weight_ih_l0",
                "decoder2.lstm.weight_hh_l0",
                "decoder2.lstm.bias_ih_l0",
                "decoder2.lstm.bias_hh_l0",
                "decoder2.fc_out.weight",
                "decoder2.fc_out.bias",
                "decoder2.attention.attn.weight",
            ]
        },
        continue_train=True,
        train_epoch=50,
        save_epoch=1,
        var_t=[
            "gpm_tp",
            "sm_surface",
            "sm_rootzone",
        ],
        var_c=[
            "area",  # 面积
            "ele_mt_smn",  # 海拔(空间平均)
            "slp_dg_sav",  # 地形坡度 (空间平均)
            "sgr_dk_sav",  # 河流坡度 (平均)
            "for_pc_sse",  # 森林覆盖率
            "glc_cl_smj",  # 土地覆盖类型
            "run_mm_syr",  # 陆面径流 (流域径流的空间平均值)
            "inu_pc_slt",  # 淹没范围 (长期最大)
            "cmi_ix_syr",  # 气候湿度指数
            "aet_mm_syr",  # 实际蒸散发 (年平均)
            "snw_pc_syr",  # 雪盖范围 (年平均)
            "swc_pc_syr",  # 土壤水含量
            "gwt_cm_sav",  # 地下水位深度
            "cly_pc_sav",  # 土壤中的黏土、粉砂、砂粒含量
            "dor_pc_pva",  # 调节程度
        ],
        var_out=["streamflow"],
        gage_id=[
            # "21401550",碧流河
            # "01181000",
            # "01411300",2021年缺失，暂时不用
            # "01414500",
            # "02016000",
            "02018000",
            # "02481510",
            # "03070500",
            # "08324000",
            # "11266500",
            # "11523200",
            # "12020000",
            # "12167000",
            # "14185000",
            # "14306500",
        ],
        which_first_tensor="batch",
        rolling=False,
        long_seq_pred=False,
        early_stopping=True,
        patience=10,
        ensemble=True,
        ensemble_items={
            "kfold": 9,
            "batch_sizes": [1024],
        },
    )
    config_data = default_config_file()
    update_cfg(config_data, args)
    return config_data


def test_dataenhanced(config):
    # train_and_evaluate(config)
    ensemble_train_and_evaluate(config)



"""
Author: Wenyu Ouyang
Date: 2021-08-09 10:19:13
LastEditTime: 2023-09-21 16:47:05
LastEditors: Wenyu Ouyang
Description: Some util functions for modeling
FilePath: /torchhydro/torchhydro/models/model_utils.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

from typing import Union
import warnings
import torch


def get_the_device(device_num: Union[list, int]):
    """
    Get device for torch according to its name

    Parameters
    ----------
    device_num : Union[list, int]
        number of the device -- -1 means "cpu" or 0, 1, ... means "cuda:x"
    """
    if device_num in [[-1], -1, ["-1"]]:
        return torch.device("cpu")
    if torch.cuda.is_available():
        return (
            torch.device(f"cuda:{str(device_num)}")
            if type(device_num) is not list
            else torch.device(f"cuda:{str(device_num[0])}")
        )
    if device_num not in [[-1], -1, ["-1"]]:
        warnings.warn("You don't have GPU, so have to choose cpu for models")
    return torch.device("cpu")
